"""Class to Manage multiple news."""

from __future__ import annotations

import asyncio
import logging
import time
from typing import TYPE_CHECKING

import pandas

from plutus_terminal.core.news.filter.filter_manager import FilterManager
from plutus_terminal.core.news.phoenix_news import PhoenixNews
from plutus_terminal.core.news.tree_news import TreeNews

if TYPE_CHECKING:
    from plutus_terminal.core.news.base import NewsFetcher, NewsMessageBus
    from plutus_terminal.core.types_ import NewsData

LOGGER = logging.getLogger(__name__)


class NewsManager:
    """Manage multiple news source."""

    def __init__(self, news_bus: NewsMessageBus) -> None:
        """Initialize shared variables.

        Args:
            news_bus (NewsMessageBus): Message bus to send news signals.
        """
        self.news_bus = news_bus
        self.news_sources: list[NewsFetcher] = [TreeNews(), PhoenixNews()]
        self._filter_manager = FilterManager()
        self._seen_links: set[str] = set()
        self._news_task: list[asyncio.Task] = []

        self.news_bus.raw_news_signal.connect(self.process_news)

    async def fetch_news(self) -> None:
        """Fetch news from news sources."""
        login_tasks = [news_fetcher.login() for news_fetcher in self.news_sources]
        await asyncio.gather(*login_tasks)
        for news_fetcher in self.news_sources:
            self._news_task.append(
                asyncio.create_task(news_fetcher.subscribe_to_wss(self.news_bus)),
            )

    async def fetch_old_news(self, limit: int) -> list[NewsData]:
        """Fetch old news from all news sources.

        Remove duplicate news from sources.

        Args:
            limit (int): Number of news to fetch.

        Returns:
            list[NewsData]: List of old news.
        """
        LOGGER.debug("Fetching old news from all news sources, limit: %s", limit)

        old_news: list[NewsData] = []
        tasks = [news_fetcher.fetch_old_news(limit) for news_fetcher in self.news_sources]
        old_news_results = await asyncio.gather(*tasks)

        # Flatten the list of lists into a single list
        old_news = [item for sublist in old_news_results for item in sublist]

        news_data_frame = pandas.DataFrame(old_news)
        news_data_frame["link"] = news_data_frame["link"].str.removesuffix("/")
        news_data_frame = news_data_frame.drop_duplicates(subset="link")
        news_data_frame = news_data_frame.sort_values(by="time", ascending=True)
        # Store displayed news to avoid duplicates
        self._seen_links.update(news_data_frame["link"])

        unique_news: list[NewsData] = news_data_frame.to_dict(orient="records")  # type: ignore

        for news in unique_news:
            self._filter_manager.filter(news)

        return unique_news[len(unique_news) - limit :]

    def process_news(self, raw_news: NewsData) -> None:
        """Process raw news.

        Validate if news is not a duplicate and add extra suggestions based on map.

        Args:
            raw_news (NewsData): News to process.
        """
        if LOGGER.isEnabledFor(logging.DEBUG):
            start_time_ms = time.time_ns() / 1000000

        # Strip trailing slash to ensre that link is not duplicated
        raw_news["link"] = raw_news["link"].removesuffix("/")

        # Check if news is already displayed based on link
        if raw_news["link"] in self._seen_links:
            LOGGER.debug("Duplicate news received: %s", raw_news["link"])
            return

        # Store displayed news to avoid duplicates
        self._seen_links.add(raw_news["link"])

        raw_news = self._filter_manager.filter(raw_news)

        if LOGGER.isEnabledFor(logging.DEBUG):
            end_time_ms = time.time_ns() / 1000000
            processed_time_ms = end_time_ms - start_time_ms  # type: ignore
            LOGGER.debug(
                "Processed message received. Process time: %f ms message: %s",
                processed_time_ms,
                raw_news,
            )

        self.news_bus.news_signal.emit(raw_news)

    async def stop_async(self) -> None:
        """Stop all async tasks and cleanup for deletion."""
        LOGGER.debug("Stopping NewsManager async")
        for news_fetcher in self.news_sources:
            await news_fetcher.stop_async()



"""Account config widgets."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from PySide6 import QtCore, QtGui, QtWidgets

from plutus_terminal.core.config import CONFIG
from plutus_terminal.ui.widgets.new_account import NewAccountDialog
from plutus_terminal.ui.widgets.toast import Toast, ToastType
from plutus_terminal.ui.widgets.top_bar_widget import TopBar

if TYPE_CHECKING:
    from plutus_terminal.core.db.models import KeyringAccount


class AccountConfig(QtWidgets.QWidget):
    """Widget to control account configs."""

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        """Initialize shared attributes."""
        super().__init__(parent=parent)

        self._main_layout = QtWidgets.QVBoxLayout()

        self._account_bar = TopBar("Manage Accounts")
        self._account_scroll_area = QtWidgets.QScrollArea()
        self._account_scroll_widget = QtWidgets.QWidget()
        self._account_box_layout = QtWidgets.QVBoxLayout()
        self._add_account_btn = QtWidgets.QPushButton("Add new Account")

        self._setup_widgets()
        self._setup_layout()
        self.populate_accounts()

    def _setup_widgets(self) -> None:
        """Config widgets."""
        self._account_scroll_area.setWidgetResizable(True)

        self._add_account_btn.setIcon(QtGui.QPixmap(":/icons/user_add"))
        self._add_account_btn.setProperty("class", "LONG")
        self._add_account_btn.setMinimumSize(80, 30)
        self._add_account_btn.clicked.connect(self._add_account)

    def _setup_layout(self) -> None:
        """Config layout."""
        self._main_layout.addWidget(self._account_bar)

        self._account_scroll_area.setWidget(self._account_scroll_widget)
        self._account_scroll_widget.setLayout(self._account_box_layout)
        self._main_layout.addWidget(self._account_scroll_area)
        account_btn_layout = QtWidgets.QHBoxLayout()
        account_btn_layout.addStretch()
        account_btn_layout.addWidget(self._add_account_btn)
        self._main_layout.addLayout(account_btn_layout)
        self._main_layout.addStretch()

        self.setLayout(self._main_layout)

    def populate_accounts(self) -> None:
        """Populate accounts."""
        # delete all account widgets from account_box_layout
        account_widgets = [
            self._account_box_layout.itemAt(index).widget()
            for index in range(self._account_box_layout.count())
            if isinstance(
                self._account_box_layout.itemAt(index).widget(),
                AccountWidget,
            )
        ]
        for widget in account_widgets:
            self._account_box_layout.removeWidget(widget)
            widget.deleteLater()

        all_accounts = CONFIG.get_all_keyring_accounts()
        for account in all_accounts:
            account_widget = AccountWidget(keyring_account=account)
            self._account_box_layout.addWidget(account_widget)
            account_widget.deleted.connect(self.populate_accounts)

    def _add_account(self) -> None:
        """Add account."""
        new_account_dialog = NewAccountDialog()
        new_account_dialog.exec()
        self.populate_accounts()
        Toast.show_message("New account added", type_=ToastType.SUCCESS)


class AccountWidget(QtWidgets.QFrame):
    """Widget to control keyring account."""

    deleted = QtCore.Signal()

    def __init__(
        self,
        keyring_account: KeyringAccount,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize shared attributes."""
        super().__init__(parent=parent)
        self._keyring_account = keyring_account

        self._main_layout = QtWidgets.QHBoxLayout()
        self._account_icon = QtWidgets.QLabel()
        self._name_label = QtWidgets.QLabel(str(self._keyring_account.username))
        self._delete_btn = QtWidgets.QPushButton()

        self._setup_widgets()
        self._setup_layout()
        self.resize(self.sizeHint())

    def _setup_widgets(self) -> None:
        """Config widgets."""
        self.setObjectName("config_item")
        self._account_icon.setPixmap(
            QtGui.QPixmap(f":/exchanges/{self._keyring_account.exchange_name}"),
        )
        self._account_icon.setScaledContents(True)
        self._account_icon.setMaximumSize(24, 24)

        self._delete_btn.setIcon(QtGui.QPixmap(":/icons/delete_icon"))
        self._delete_btn.setIconSize(QtCore.QSize(24, 24))
        self._delete_btn.setProperty("class", "borderless")
        self._delete_btn.clicked.connect(self._delete_account)

    def _setup_layout(self) -> None:
        """Config layout."""
        self._main_layout.addWidget(self._account_icon)
        self._main_layout.addWidget(self._name_label)
        self._main_layout.addStretch()
        self._main_layout.addWidget(self._delete_btn)
        self.setLayout(self._main_layout)

    def _delete_account(self) -> None:
        """Delete account."""
        CONFIG.delete_account(self._keyring_account.id)  # type: ignore
        self.deleted.emit()
        Toast.show_message(
            f"Account '{self._keyring_account.username}' deleted",
            type_=ToastType.SUCCESS,
        )



"""Base class for Exchange."""

from __future__ import annotations

from abc import ABC, abstractmethod
import asyncio
from decimal import Decimal
import logging
import time
from typing import TYPE_CHECKING, Optional, Protocol, Self

from PySide6.QtCore import QObject, Signal
from qasync import asyncSlot

from plutus_terminal.core.config import CONFIG
from plutus_terminal.core.exceptions import (
    OptionsNotAvailableError,
    TransactionFailedError,
)
from plutus_terminal.core.exchange.types import PnlDetails
from plutus_terminal.core.types_ import (
    ExchangeType,
    NewAccountInfo,
    OptionsDirection,
    PerpsPosition,
    PerpsTradeType,
    PriceData,
    PriceHistory,
)
from plutus_terminal.ui.widgets.toast import Toast, ToastType

if TYPE_CHECKING:
    import pandas

    from plutus_terminal.core.exchange.types import OrderData, TradeResults
    from plutus_terminal.core.types_ import (
        PerpsTradeDirection,
    )

LOGGER = logging.getLogger(__name__)


class ExchangeFetcherMessageBus(QObject):
    """Message Bus for all fetch related events."""

    price_history_signal = Signal(PriceHistory)
    subscribed_prices_signal = Signal(dict)
    balance_signal = Signal(Decimal)
    positions_signal = Signal(list)  # list[PerpsPosition]
    orders_signal = Signal(list)  # list[OrderData]
    price_synced = Signal(bool)


class ExchangeFetcher(Protocol):
    """Protocol for exchange fetch information."""

    async def fetch_price_history(
        self,
        pair: str,
        from_timestamp: int,
        resolution: str,
    ) -> PriceHistory:
        """Fetch the price history of pair for the given start and end time.

        Args:
            pair (str): Pair to fetch.
            from_timestamp (int): Timestamp for the leftmost bar.
            resolution (str): Bar resolution in seconds.

        Returns:
            PriceHistory: Dict with ohlcv.
        """
        ...

    async def subscribe_to_price(self, pair: str, force: bool = False) -> None:
        """Subscribe to receive price updates for the given pair."""
        ...

    async def resubscribe_on_going_connections(self) -> None:
        """Resubscribe on going connections."""
        ...

    async def unsubscribe_to_price(self, pair: str, force: bool = False) -> None:
        """Unsubscribe to receive price updates for the given pair."""
        ...

    async def receive_subscribed_prices(self) -> None:
        """Receive live data of subscribed prices from exchange."""
        ...

    async def watch_all_positions(self) -> None:
        """Watch all open positions."""
        ...

    async def fetch_all_positions(self) -> list[PerpsPosition]:
        """Fetch all open positions."""
        ...

    async def watch_all_orders(self) -> None:
        """Watch all open orders."""
        ...

    async def fetch_all_orders(self) -> list[OrderData]:
        """Fetch all open orders."""
        ...

    async def fetch_price_at_time(self, pair: str, timestamp: int) -> PriceData:
        """Fetch price at current time."""
        ...

    async def fetch_current_price(self, pair: str) -> PriceData:
        """Fetch current price of given pair."""
        ...

    async def watch_stable_balance(self) -> None:
        """Watch stable balance."""
        ...

    async def fetch_stable_balance(self) -> Decimal:
        """Fetch stable balance."""
        ...

    def get_position_associated_with_order(self, order: OrderData) -> Optional[PerpsPosition]:
        """Get position associated with given order.

        Args:
            order (OrderData): Order to get position for.

        Returns:
            Optional[PerpsPosition]: Position associated with given order.
        """
        ...

    def calculate_margin_fee(self, position_size: Decimal) -> Decimal:
        """Calcualte margin fee for a given position.

        Args:
            position_size (Decimal): Position size to get fee for.

        Returns:
            Decimal: Fee amount in USD Stable Format.
        """
        ...

    def fetch_funding_fee(self, perps_position: PerpsPosition) -> Decimal:
        """Fetch funding fee for a given position.

        Args:
            perps_position (PerpsPosition): Position to get funding fee for.

        Returns:
            Decimal: Funding fee in USD Stable Format.
        """
        ...

    def calculate_liquidation_price(self, perps_position: PerpsPosition) -> Decimal:
        """Calculate liquidation price for a given position.

        Args:
            perps_position (PerpsPosition): Position to get liquidation price for.

        Returns:
            Decimal: Liquidation price in USD Stable Format.
        """
        ...

    def calculate_pnl_percent_before_fees(
        self,
        perps_position: PerpsPosition,
        current_price: Optional[Decimal],
    ) -> Decimal:
        """Calculate pnl percent for a given position.

        Args:
            perps_position (PerpsPosition): Position to get pnl for.
            current_price (Optional[Decimal]): Current price of the pair.

        Returns:
            Decimal: PNL in USD Stable Format.
        """
        ...

    async def stop_async(self) -> None:
        """Stop infinite loops and close connections."""
        ...


class ExchangeTrader(Protocol):
    """Protocol for exchange trading actions."""

    async def create_order(self, trade_arguments) -> TradeResults:  # noqa: ANN001
        """Create new order.

        Args:
            trade_arguments (dict): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.

        Returns:
            TradeResults: Result of the trade.
        """
        ...

    async def create_reduce_order(self, trade_arguments) -> TradeResults:  # noqa: ANN001
        """Create new reduce only order.

        Args:
            trade_arguments (dict): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.

        Raises:
            NotImplementedError: If the trade direction is not supported.

        Returns:
            TradeResults: Result of the trade.
        """
        ...

    async def close_position(self, trade_arguments) -> TradeResults:  # noqa: ANN001
        """Close the give position.

        This are for positions that are already bought.

        Args:
            trade_arguments (dict): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.

        Returns:
            TradeResults: Result of the trade.
        """
        ...

    async def cancel_order(self, trade_arguments) -> TradeResults:  # noqa: ANN001
        """Cancel the given order.

        Args:
            trade_arguments (dict): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.

        Returns:
            TradeResults: Result of the trade.
        """
        ...

    async def edit_order(
        self,
        trade_arguments,  # noqa: ANN001
    ) -> TradeResults:
        """Update the given order.

        Args:
            trade_arguments (dict): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.

        Returns:
            TradeResults: Result of the trade.
        """
        ...


class ExchangeTraderDex(ExchangeTrader, Protocol):
    """Protocol for dex trading actions."""

    async def approve_contracts(self) -> None:
        """Approve necessary contracts for trading."""
        ...


class ExchangeOptions(Protocol):
    """Protocol for exchange options actions."""

    async def fetch_orders(self, options_orders_params) -> pandas.DataFrame:  # noqa: ANN001
        """Fetch options orders based on parameters.

        Args:
            options_orders_params (OrdersParams): Params for request.

        Returns:
            pandas.DataFrame: DataFrame with options orders.
        """
        ...

    async def buy_options(self, options_buy_params) -> None:  # noqa: ANN001
        """Buy options.

        Args:
            options_buy_params: Params to buy options.
        """
        ...

    def fetch_available_pairs(self) -> list[str]:
        """Fetch available pairs for options."""
        ...

    async def filter_with_strategy(
        self,
        direction: OptionsDirection,
        value: Decimal,
        pair: str,
    ) -> pandas.DataFrame:
        """Filter options based on strategy.

        Args:
            direction (OptionsDirection): Direction to buy the option.
            value (Decimal): Max value to spend.
            pair (str): Pair to open options for.

        Returns:
            pandas.DataFrame: DataFrame with options orders.
        """
        ...


class ExchangeBase(ABC):
    """Base class to interact with exchange."""

    def __init__(self, fetcher_bus: ExchangeFetcherMessageBus) -> None:
        """Initialize shared variables."""
        self.fetcher_bus = fetcher_bus
        self._is_price_synced = True
        self._watched_positions: list[PerpsPosition] = []
        self._async_tasks: list[asyncio.Task] = []

    @property
    @abstractmethod
    def trader(self) -> ExchangeTrader:
        """Returns: Exchange Trader."""

    @property
    @abstractmethod
    def fetcher(self) -> ExchangeFetcher:
        """Returns: Exchange Trader."""

    @property
    @abstractmethod
    def available_pairs(self) -> set:
        """Returns set with all available pairs."""

    @property
    @abstractmethod
    def pair_prefix(self) -> str:
        """Returns pair prefix."""

    @property
    @abstractmethod
    def pair_separator(self) -> str:
        """Returns pair separator."""

    @property
    @abstractmethod
    def quote_symbol(self) -> str:
        """Returns quote symbol."""

    @property
    @abstractmethod
    def pair_suffix(self) -> str:
        """Returns pair suffix."""

    @property
    @abstractmethod
    def cached_prices(self) -> dict:
        """Return all prices cache."""

    @property
    @abstractmethod
    def stable_balance(self) -> Decimal:
        """Return stable balance."""

    @property
    def account_info(self) -> dict[str, str]:
        """Return info to be added to account info widget."""
        return {
            "Exchange": self.name(),
        }

    @classmethod
    @abstractmethod
    async def create(cls, fetcher_bus: ExchangeFetcherMessageBus) -> Self:
        """Create class instance and init_async."""

    @property
    def default_pair(self) -> str:
        """Return default pair name to auto subscribe."""
        return self.format_pair_from_coin("BTC")

    @property
    def min_order_size(self) -> Decimal:
        """Return min trade size."""
        return Decimal(0)

    @property
    def max_order_size(self) -> Decimal:
        """Return max trade size."""
        return Decimal(100_000_000_000_000_000_000_000)

    @property
    def options(self) -> ExchangeOptions:
        """Returns: Exchange Options."""
        raise OptionsNotAvailableError

    def has_options(self) -> bool:
        """Return if exchange has options."""
        try:
            return self.options is not None
        except OptionsNotAvailableError:
            return False

    @abstractmethod
    async def is_ready_to_trade(self) -> bool:
        """Check if account is ready to trade.

        Mainly used for DEXs to check if contract appovals are done.

        Returns:
            bool: True if account is ready to trade.
        """

    @abstractmethod
    async def approve_for_trading(self) -> None:
        """Approve contracts needed for trading.

        Mainly used for DEXs to approve contract appovals.
        """

    def format_pair_from_coin(self, coin: str) -> str:
        """Format coin to pair name."""
        return f"{self.pair_prefix}{coin}{self.pair_separator}{self.quote_symbol}{self.pair_suffix}"

    def format_simple_pair_from_pair(self, pair: str) -> str:
        """Format pair name to simple pair."""
        simple_pair = ""
        if self.pair_prefix:
            simple_pair = pair[len(self.pair_prefix) :]

        if self.pair_suffix:
            simple_pair = simple_pair[: -len(self.pair_suffix)]
        return simple_pair

    def format_coin_from_pair(self, pair: str) -> str:
        """Format pair name to coin."""
        simple_pair = self.format_simple_pair_from_pair(pair)
        return simple_pair.split(self.pair_separator)[0]

    async def fetch_prices(self) -> None:
        """Fetch prices in an infinite loop."""
        await self.fetcher.subscribe_to_price(self.default_pair)
        self._async_tasks.append(
            asyncio.create_task(self.fetcher.receive_subscribed_prices()),
        )
        self._async_tasks.append(
            asyncio.create_task(self.fetcher.watch_all_positions()),
        )
        self._async_tasks.append(asyncio.create_task(self.fetcher.watch_all_orders()))
        self._async_tasks.append(asyncio.create_task(self.fetcher.watch_stable_balance()))
        self.fetcher_bus.positions_signal.connect(self._update_watched_positions)

    @asyncSlot()
    async def _update_watched_positions(
        self,
        all_positions: list[PerpsPosition],
    ) -> None:
        """Update list of current watched positions.

        Subscribe or remove subscriptions based on new positions.
        """
        if all_positions != self._watched_positions:
            to_subscribe = [
                position["pair"]
                for position in all_positions
                if position not in self._watched_positions
            ]
            to_unsubscribe = [
                position["pair"]
                for position in self._watched_positions
                if position not in all_positions
            ]
            for pair in to_subscribe:
                await self.fetcher.subscribe_to_price(pair)

            for pair in to_unsubscribe:
                await self.fetcher.unsubscribe_to_price(pair)

            self._watched_positions = all_positions

    async def fetch_price_history(
        self,
        pair: str,
        resolution: str,
        bars_num: int = 200,
    ) -> PriceHistory:
        """Fetch current price history for current pair.

        Args:
            pair (str): Pair to fetch.
            resolution (str): Bar resolution.
            bars_num (int): Number of bars expected for resolution.
        """
        now_timestamp = int(time.time())
        match resolution:
            case "5":
                start_timestamp = now_timestamp - 5 * 60 * bars_num
            case "15":
                start_timestamp = now_timestamp - 15 * 60 * bars_num
            case "30":
                start_timestamp = now_timestamp - 30 * 60 * bars_num
            case "60":
                start_timestamp = now_timestamp - 60 * 60 * bars_num
            case "240":
                start_timestamp = now_timestamp - 240 * 60 * bars_num
            case _:
                start_timestamp = now_timestamp - bars_num * 60

        return await self.fetcher.fetch_price_history(
            pair=pair,
            from_timestamp=start_timestamp,
            resolution=resolution,
        )

    @asyncSlot()
    async def set_all_leverage(self, leverage: int) -> None:
        """Set leverage for all positions.

        Args:
            leverage (int): Leverage to set.
        """
        Toast.show_message(
            f"Leverage set to all pairs: {leverage}x",
            type_=ToastType.SUCCESS,
        )
        CONFIG.leverage = leverage

    @asyncSlot()
    async def set_leverage(self, coin: str, leverage: int) -> None:
        """Set leverage for pair.

        Args:
            coin (str): Coin to set leverage for.
            leverage (int): Leverage to set.
        """
        pair = self.format_pair_from_coin(coin)
        Toast.show_message(
            f"Leverage of {pair} set to: {leverage}x",
            type_=ToastType.SUCCESS,
        )
        CONFIG.leverage = leverage

    def is_valid_order_size(self, order_size: Decimal) -> bool:
        """Validate order size with min and max values.

        Args:
            order_size (Decimal): Size to validate.

        Returns:
            bool: True if order size is valid.
        """
        if order_size < self.min_order_size or order_size > self.max_order_size:
            return False
        return True

    @abstractmethod
    async def create_order(
        self,
        pair: str,
        amount: Decimal,
        trade_direction: PerpsTradeDirection,
        trade_type: PerpsTradeType,
        execution_price: Optional[Decimal] = None,
        take_profit: Optional[Decimal] = None,
        stop_loss: Optional[Decimal] = None,
    ) -> None:
        """Create new order.

        Args:
            pair (str): Pair to open trade for.
            amount (Decimal): Value in stable to open trade for, this will be multiplied
                by de configured leverage.
            trade_direction (TradeDirection): Trade direction.
            trade_type (TradeType): Trade type.
            execution_price (Optional[Decimal], optional): Execution price.
            take_profit (Optional[float], optional): Take profit price.
                If None use CONFIG.take_profit.
            stop_loss (Optional[float], optional): Stop loss price.
                If None use CONFIG.stop_loss.

        Raises:
            InvalidOrderSizeError: If order size is not valid.
            TransactionFailedError: If transaction failed
        """

    @abstractmethod
    async def edit_order(
        self,
        order_data: OrderData,
        new_size_stable: Decimal,
        new_execution_price: Optional[Decimal],
    ) -> None:
        """Edit existent order.

        Args:
            order_data (OrderData): Order to edit.
            new_size_stable (Decimal): Value in stable to open trade for, this will be multiplied
                by de configured leverage.
            new_execution_price (Optional[Decimal], optional): Execution price.

        Raises:
            TransactionFailedError: If transaction failed
        """
        ...

    @abstractmethod
    async def create_reduce_order(
        self,
        pair: str,
        size: Decimal,
        trade_direction: PerpsTradeDirection,
        trade_type: PerpsTradeType,
        execution_price: Optional[Decimal],
    ) -> None:
        """Create new reduce only order.

        Args:
            pair (str):  Pair to open trade for.
            size (Decimal): Value in stable to open trade for, this will be multiplied
                by de configured leverage.
            trade_direction (TradeDirection): Trade direction.
            trade_type (TradeType): Trade type.
            execution_price (Decimal): Execution price.

        Raises:
            TransactionFailedError: If transaction failed
        """

    @abstractmethod
    async def cancel_order(self, order_data: OrderData) -> None:
        """Cancel the given order.

        Args:
            order_data (OrderData): Order to close.

        Raises:
            TransactionFailed: If the transaction fails.
        """
        ...

    @asyncSlot()
    async def close_position(self, perps_position: PerpsPosition) -> None:
        """Close position for pair.

        Close position and emit positions signal.

        Args:
            perps_position (PerpsPosition): Position to close.

        Raises:
            TransactionFailedError: If transaction failed
        """
        toast_id = Toast.show_message(
            f"Closing position for {perps_position['pair']}",
            type_=ToastType.WARNING,
        )
        try:
            await self.trader.close_position(perps_position)
        except TransactionFailedError as error:
            Toast.update_message(
                toast_id,
                f"Failed to close position > {error}",
                type_=ToastType.ERROR,
            )
        all_positions = await self.fetcher.fetch_all_positions()
        self.fetcher_bus.positions_signal.emit(all_positions)
        Toast.update_message(toast_id, "Position closed", type_=ToastType.SUCCESS)

    def get_position_associated_with_order(self, order: OrderData) -> Optional[PerpsPosition]:
        """Get position associated with given order.

        Args:
            order (OrderData): Order to get position for.

        Returns:
            Optional[PerpsPosition]: Position associated with given order.
        """
        return self.fetcher.get_position_associated_with_order(order)

    def calculate_margin_fee(self, position_size: Decimal) -> Decimal:
        """Caculate margin fee for a given position.

        Args:
            position_size (Decimal): Position size to get fee for.

        Returns:
            Decimal: Position fee.
        """
        return self.fetcher.calculate_margin_fee(position_size)

    def fetch_funding_fee(self, perps_position: PerpsPosition) -> Decimal:
        """Fetch funding fee for a given position.

        Args:
            perps_position (PerpsPosition): Position to get funding fee for.

        Returns:
            Decimal: Funding fee in USD Stable Format.
        """
        return self.fetcher.fetch_funding_fee(perps_position)

    def calculate_liquidation_price(self, perps_position: PerpsPosition) -> Decimal:
        """Calculate liquidation price for a given position.

        Args:
            perps_position (PerpsPosition): Position to get liquidation price for.

        Returns:
            Decimal: Liquidation price in USD Stable Format.
        """
        return self.fetcher.calculate_liquidation_price(perps_position)

    def calculate_pnl(
        self,
        perps_position: PerpsPosition,
        current_price: Optional[Decimal],
    ) -> PnlDetails:
        """Calculate pnl for a given position.

        Args:
            perps_position (PerpsPosition): Position to get pnl for.
            current_price (Optional[Decimal]): Current price of the pair. If None, use last price

        Returns:
            Decimal: PNL in USD Stable Format.
        """
        leverage = perps_position["leverage"]
        trade_collateral = perps_position["position_size_stable"] / leverage
        position_fee = self.fetcher.calculate_margin_fee(perps_position["position_size_stable"])
        funding_fee = self.fetcher.fetch_funding_fee(perps_position)
        pnl_percentage = self.fetcher.calculate_pnl_percent_before_fees(
            perps_position,
            current_price,
        )
        pnl_usd = (trade_collateral * pnl_percentage) / 100
        pnl_usd_after_fees = pnl_usd - (2 * position_fee) - funding_fee
        pnl_percentage_after_fees = pnl_usd_after_fees * 100 / trade_collateral

        return PnlDetails(
            {
                "pnl_usd_before_fees": pnl_usd,
                "pnl_percentage_before_fees": pnl_percentage,
                "position_fee_usd": position_fee,
                "funding_fee_usd": funding_fee,
                "pnl_usd_after_fees": pnl_usd_after_fees,
                "pnl_percentage_after_fees": pnl_percentage_after_fees,
            },
        )

    async def buy_options_with_strategy(
        self,
        direction: OptionsDirection,  # noqa: ARG002
        value: Decimal,  # noqa: ARG002
        pair: str,  # noqa: ARG002
    ) -> None:
        """Buy options based on strategy.

        Args:
            direction (OptionsDirection): Direction to buy the option.
            value (Decimal): Max value in stable to spend.
            pair (str): Pair to open options for.
        """
        raise OptionsNotAvailableError

    async def stop_async(self) -> None:
        """Stop all async tasks and cleanup for deletion."""
        LOGGER.debug("Stopping exchange: `%s`", self.name())
        for task in self._async_tasks:
            task.cancel()
        await self.fetcher.stop_async()

    @staticmethod
    @abstractmethod
    def name() -> str:
        """Return exchange name."""

    @staticmethod
    @abstractmethod
    def new_account_info() -> NewAccountInfo:
        """Provide info for new account creation."""

    @staticmethod
    @abstractmethod
    def exchange_type() -> ExchangeType:
        """Return exchange type."""

    @staticmethod
    @abstractmethod
    def validate_secrets(secrets: list[str]) -> tuple[bool, str]:
        """Validate secrets."""



"""News filters actions."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Protocol

from plutus_terminal.core.news.filter.types import ActionType

if TYPE_CHECKING:
    from textsearch import TSResult

    from plutus_terminal.core.types_ import NewsData


class FilterAction(Protocol):
    """Protocol for news filter actions."""

    def __call__(
        self,
        news_data: NewsData,
        text_data_key: str,
        search_result: Optional[TSResult],
        **kwargs: dict,
    ) -> Any: ...  # noqa: ANN401


def coin_association_action(
    news_data: NewsData,
    text_data_key: str,
    search_result: Optional[TSResult],
    **kwargs: dict[Any, Any],
) -> NewsData:
    """Associate coin with news if found.

    Add coin symbol to news data and modify text to be shown with color if
    text_data_key is provided.

    Args:
        news_data (NewsData): News data to be modified.
        text_data_key (str): Key in news data of text being modified.
        search_result (Optional[TSResult]): Text search results used for text substitution.
        **kwargs: Keyword arguments. Valid arguments are:
            * coin (str): Coin symbol
            * color (tuple[int, int, int]): RGB color for text.
    """
    news_data["coin"].add(kwargs["coin"])
    # Only replace text if this key is provided
    if text_data_key and search_result:
        text = news_data[text_data_key]  # type: ignore
        # Ensure color keyword is a tuple
        text = (
            f"{text[:search_result.start]}<span style='color: rgb{tuple(kwargs['color'])};'>"
            f"{search_result.match}</span>{text[search_result.end:]}"
        )
        news_data[text_data_key] = text  # type: ignore
    return news_data


def sound_association_action(
    news_data: NewsData,
    text_data_key: str,
    search_result: Optional[TSResult],
    **kwargs: dict[Any, Any],
) -> NewsData:
    """Associate sound with news if found.

    Add sound to news data and modify text to be shown with color if
    text_data_key is provided.

    Args:
        news_data (NewsData): News data to be modified.
        text_data_key (str): Key in news data of text being modified.
        search_result (Optional[TSResult]): Text search results used for text substitution.
        text_part (str): Text part to be modified.
        **kwargs: Keyword arguments. Valid arguments are:
            * sound_path (str): QResources path to sound.
            * color (tuple[int, int, int]): RGB color for text.
    """
    news_data["sfx"] = kwargs["sound_path"]
    # Only replace text if this key is provided
    if text_data_key and search_result:
        text = news_data[text_data_key]  # type: ignore
        # Ensure color keyword is a tuple
        text = (
            f"{text[:search_result.start]}<span style='color: rgb{tuple(kwargs['color'])};'>"
            f"{search_result.match}</span>{text[search_result.end:]}"
        )
        news_data[text_data_key] = text  # type: ignore
    return news_data


def ignore_action(
    news_data: NewsData,
    text_data_key: str,  # noqa: ARG001
    search_result: Optional[TSResult],  # noqa: ARG001
    **kwargs: dict[Any, Any],  # noqa: ARG001
) -> NewsData:
    """Ignore news.

    Args:
        news_data (NewsData): News data to be ignored.
        text_data_key (str): Argument not used!
        search_result (Optional[TSResult]): Argument not used!
        **kwargs: Argument not used!
    """
    news_data["ignored"] = True
    return news_data


FILTER_ACTIONS_MAP = {
    ActionType.COIN_ASSOCIATION: coin_association_action,
    ActionType.SOUND_ASSOCIATION: sound_association_action,
    ActionType.IGNORE: ignore_action,
}



"""Widget to visualize open positions."""

from __future__ import annotations

from copy import deepcopy
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Optional

from PySide6.QtCore import (
    QAbstractItemModel,
    QAbstractTableModel,
    QModelIndex,
    QPoint,
    Qt,
    Signal,
)
from PySide6.QtGui import QBrush, QColor, QFont, QIcon
from PySide6.QtWidgets import (
    QAbstractItemView,
    QButtonGroup,
    QGridLayout,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QMenu,
    QPushButton,
    QRadioButton,
    QTableView,
    QWidget,
    QWidgetAction,
)
from qasync import asyncSlot

from plutus_terminal.core.exchange.types import OrderData, PerpsTradeType, PriceData
from plutus_terminal.core.types_ import PerpsPosition, PerpsTradeDirection
from plutus_terminal.ui import ui_utils
from plutus_terminal.ui.widgets.decimal_spin_box import (
    DecimalSpinBox,
)
from plutus_terminal.ui.widgets.manage_order import ManageOrder
from plutus_terminal.ui.widgets.pnl_breakdown import PnlBreakdown

if TYPE_CHECKING:
    from collections.abc import Callable

    from plutus_terminal.core.exchange.base import ExchangeBase

HEADER_MAP = {
    "pair": "Pair",
    "trade_direction": "Side",
    "collateral_stable": "Collateral",
    "leverage": "Lev",
    "position_size_stable": "Size",
    "open_price": "Avg Entry",
    "liquidation_price": "Est. Liq. Price",
    "pnl": "PnL",
    "close": "Close",
}

FILL_LATER = {"close", "pnl"}


class PositionsTableModel(QAbstractTableModel):
    """Table Model to display open positions."""

    def __init__(
        self,
        format_simple_pair: Callable[[str], str],
        data: Optional[list[PerpsPosition]] = None,
    ) -> None:
        """Initialize shared variables."""
        super().__init__()
        self._data = data if data else []
        self.headers_source = list(HEADER_MAP.keys())
        self.format_simple_pair = format_simple_pair

    def data(self, index, role: Qt.ItemDataRole) -> Any:  # noqa: C901, PLR0911
        """Define how data is displayed."""
        current_header = self.headers_source[index.column()]

        if current_header not in FILL_LATER:
            value = self._data[index.row()][current_header]  # type: ignore
        else:
            value = None

        if role == Qt.ItemDataRole.DisplayRole:
            if value is None:
                return None
            if current_header == "trade_direction":
                return value.name
            if current_header == "leverage":
                return f"{value}x"
            if current_header == "pair":
                # Removing the Pair prefix if any
                return self.format_simple_pair(value)
            if isinstance(value, Decimal):
                minimal_digits = ui_utils.get_minimal_digits(float(value), 3)
                return f"${value:,.{minimal_digits}f}"
            return value

        if role == Qt.ItemDataRole.ForegroundRole:
            if current_header == "trade_direction":
                if value is PerpsTradeDirection.LONG:
                    return QBrush(QColor("green"))
                return QBrush(QColor("red"))
            if current_header == "liquidation_price":
                return QBrush(QColor("orange"))

        if role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignCenter

        if role == Qt.ItemDataRole.FontRole and current_header == "trade_direction":
            font = QFont()
            font.setBold(True)
            return font

        if role == Qt.ItemDataRole.UserRole:
            return self._data[index.row()]
        return None

    def headerData(self, section, orientation, role):
        """Define header data."""
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return HEADER_MAP[self.headers_source[section]]
            return None
        return None

    def rowCount(self, index=None):
        """Define row count."""
        return len(self._data)

    def columnCount(self, index=None):
        """Define column count."""
        return len(HEADER_MAP)

    def update_positions(self, data: list[PerpsPosition]) -> None:
        """Update open positions."""
        # Only reset model if data changed.
        if data == self._data:
            return
        self.beginResetModel()
        self._data = data
        self.endResetModel()

    def on_new_exchange(self, new_exchange: ExchangeBase) -> None:
        """Update info based on new exchange.

        Args:
            new_exchange (ExchangeBase): New exchangeBase.
        """
        self.format_simple_pair = new_exchange.format_simple_pair_from_pair


class PositionsTableView(QTableView):
    """Table view to display open positions."""

    row_clicked = Signal(str)

    def __init__(
        self,
        exchange: ExchangeBase,
        parent: Optional[QWidget] = None,
    ) -> None:
        """Initialize shared attributes."""
        super().__init__(parent=parent)
        self._exchange = exchange
        self._close_index = list(HEADER_MAP).index("close")
        self._pnl_index = list(HEADER_MAP).index("pnl")
        self._pnl_widgets: dict[str, dict[bool, PnlBreakdown]] = {}
        self._position_manager_widgets: dict[str, dict[bool, PositionManager]] = {}
        self._cached_prices: dict[str, PriceData] = {}
        self.clicked.connect(self.on_row_click)
        self._setup_style()

    def _setup_style(self) -> None:
        """Set Table style."""
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.Stretch)
        self.setAlternatingRowColors(True)
        self.verticalHeader().setVisible(False)
        self.setSelectionMode(QAbstractItemView.SelectionMode.NoSelection)
        self.setShowGrid(False)

    def setModel(self, model: QAbstractItemModel) -> None:
        """Override setModel to add close buttons."""
        super().setModel(model)
        model.modelReset.connect(self.add_position_manager)
        model.modelReset.connect(self.create_pnl_breakdowns)

    def add_position_manager(self) -> None:
        """Add position manager for each row."""
        self.blockSignals(True)
        for row in range(self.model().rowCount()):
            index = self.model().index(row, self._close_index)
            data = index.data(Qt.ItemDataRole.UserRole)

            position_manager = ui_utils.create_stored_widget(
                PositionManager,
                self._position_manager_widgets,
                data["pair"],
                data["trade_direction"],
                position=data,
                exchange=self._exchange,
                parent=self,
            )

            self.setIndexWidget(index, position_manager)
        self.horizontalHeader().setSectionResizeMode(
            self._close_index,
            QHeaderView.ResizeMode.Fixed,
        )

        widget = self.indexWidget(self.model().index(0, self._close_index))
        if widget is not None:
            self.setColumnWidth(
                self._close_index,
                int(widget.sizeHint().width() * 1.05),
            )
        self.blockSignals(False)

    def update_cached_prices(self, cached_prices: dict[str, PriceData]) -> None:
        """Update cached prices."""
        self._cached_prices = cached_prices
        self.update_pnl(cached_prices)

    def create_pnl_breakdowns(self) -> None:
        """Create pnl breakdowns for each row."""
        self.blockSignals(True)
        for row in range(self.model().rowCount()):
            data = self.model().index(row, 0).data(Qt.ItemDataRole.UserRole)
            ui_utils.create_stored_widget(
                PnlBreakdown,
                self._pnl_widgets,
                data["pair"],
                data["trade_direction"],
            )

        self.blockSignals(False)

    def update_pnl(self, cached_prices: dict[str, PriceData]) -> None:
        """Update pries for open positions."""
        for row in range(self.model().rowCount()):
            data = self.model().index(row, 0).data(Qt.ItemDataRole.UserRole)
            try:
                current_price = cached_prices[data["pair"]]["price"]
            except KeyError:
                return

            pnl_index = self.model().index(row, self._pnl_index)

            pnl_details = self._exchange.calculate_pnl(data, current_price)

            trade_direction = data["trade_direction"]
            pnl_widget = ui_utils.get_stored_widget(
                self._pnl_widgets,
                data["pair"],
                trade_direction,
            )
            if pnl_widget is None or not isinstance(pnl_widget, PnlBreakdown):
                return

            pnl_widget.set_pnl(
                pnl_details["pnl_usd_after_fees"],
                pnl_details["pnl_percentage_after_fees"],
            )
            pnl_widget.set_tooltip_content(
                pnl_details["pnl_usd_before_fees"],
                pnl_details["funding_fee_usd"],
                pnl_details["position_fee_usd"],
                pnl_details["pnl_usd_after_fees"],
            )
            self.setIndexWidget(pnl_index, pnl_widget)
            # Set row height to 2x to make it fit in the table
            self.setRowHeight(row, int(self.sizeHintForRow(row) * 2))
        self.horizontalHeader().setSectionResizeMode(
            self._pnl_index,
            QHeaderView.ResizeMode.Fixed,
        )

        widget = self.indexWidget(self.model().index(0, self._pnl_index))
        if widget is not None:
            self.setColumnWidth(self._pnl_index, int(widget.sizeHint().width() * 1.2))

    def on_row_click(self, index: QModelIndex) -> None:
        """Handle click on row."""
        self.row_clicked.emit(index.data(Qt.ItemDataRole.UserRole)["pair"])

    def on_new_exchange(self, new_exchange: ExchangeBase) -> None:
        """Update info based on new exchange.

        Args:
            new_exchange (ExchangeBase): New exchangeBase.
        """
        self._exchange = new_exchange
        self._position_manager_widgets = {}
        self._pnl_widgets = {}


class PositionManager(QWidget):
    """Position manager widget."""

    def __init__(
        self,
        position: PerpsPosition,
        exchange: ExchangeBase,
        parent: Optional[QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(parent)
        self._position = position
        self._exchange = exchange

        self._main_layout = QHBoxLayout()
        self.close_button = QPushButton("Close")
        self.tp_sl_button = QPushButton("TP/SL")

        self._menu = QMenu(self)
        self._close_action = PositionCloseAction(self._position, self)

        self._setup_widgets()
        self._setup_layout()

    def _setup_widgets(self) -> None:
        """Condifure widgets."""
        self._menu.setObjectName("floating")
        self._close_action.reduce_clicked.connect(self._on_close_reduce_clicked)
        self._close_action.set_price_clicked.connect(self.set_price_limit)
        self._menu.addAction(self._close_action)

        self.close_button.setProperty("class", "gray")
        self.close_button.setToolTip("Close Position. Right click for more options.")
        self.close_button.setMinimumSize(50, 30)
        self.close_button.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.close_button.customContextMenuRequested.connect(self.show_close_context)
        self.close_button.clicked.connect(self.close_position)

        self.tp_sl_button.setProperty("class", "gray")
        self.tp_sl_button.setToolTip("Open dialog to set TP/SL.")
        self.tp_sl_button.setMinimumSize(50, 30)
        self.tp_sl_button.clicked.connect(self.on_tp_sl_clicked)

    def _setup_layout(self) -> None:
        """Organize layouts."""
        self._main_layout.setContentsMargins(0, 0, 0, 0)
        self._main_layout.addWidget(self.close_button)
        self._main_layout.addWidget(self.tp_sl_button)
        self.setLayout(self._main_layout)

    def show_close_context(self, pos: QPoint) -> None:
        """Show context menu."""
        self._menu.exec_(
            self.close_button.mapToGlobal(self.close_button.rect().center()),
        )

    def set_price_limit(self) -> None:
        """Set price for limit trade."""
        self._close_action.set_price_limit(
            self._exchange.cached_prices[self._position["pair"]]["price"],
        )

    @asyncSlot()
    async def close_position(self) -> None:
        """Close position."""
        await self._exchange.close_position(self._position)

    @asyncSlot()
    async def _on_close_reduce_clicked(self, kwargs: dict) -> None:
        """Handle click on reduce."""
        # Close position if size matches
        if self._position["position_size_stable"] == kwargs["size"]:
            await self._exchange.close_position(self._position)
            return
        await self._exchange.create_reduce_order(**kwargs)

    def on_tp_sl_clicked(self) -> None:
        """Handle click on TP/SL."""
        order_dialog = ManageOrder(
            OrderData(
                {
                    "pair": self._position["pair"],
                    "order_type": PerpsTradeType.TRIGGER_TP,
                    "size_stable": self._position["position_size_stable"],
                    "trigger_price": self._position["open_price"],
                    "id": "_",
                    "reduce_only": True,
                    "trade_direction": self._position["trade_direction"],
                },
            ),
            exchange=self._exchange,
            associated_position=deepcopy(self._position),
            parent=self,
        )
        order_dialog.execute_order.connect(self._handle_tp_sl_clicked)
        order_dialog.show()

    @asyncSlot()
    async def _handle_tp_sl_clicked(self, order_data: OrderData) -> None:
        """Execute order on exchange.

        Args:
            order_data (OrderData): Order to execute.
        """
        await self._exchange.create_reduce_order(
            pair=order_data["pair"],
            size=order_data["size_stable"],
            trade_direction=order_data["trade_direction"],
            trade_type=order_data["order_type"],
            execution_price=order_data["trigger_price"],
        )


class PositionCloseAction(QWidgetAction):
    """Widget to close position.."""

    reduce_clicked = Signal(dict)
    set_price_clicked = Signal()

    def __init__(self, position: PerpsPosition, parent: Optional[QWidget] = None) -> None:
        """Initialize widget."""
        super().__init__(parent)  # type: ignore
        self._position = position

        self._default_widget = QWidget(parent)
        self._main_layout = QGridLayout()
        self.type_group = QButtonGroup()
        self.market_button = QRadioButton("Market")
        self.limit_button = QRadioButton("Limit")
        self.type_group_layout = QHBoxLayout()

        self._amount_label = QLabel("Amount:")
        self._amount_box = DecimalSpinBox()
        self._amount_group = QButtonGroup()
        self._amount_group_layout = QHBoxLayout()
        for value in ("25%", "50%", "75%", "100%"):
            button = QRadioButton(value)
            self._amount_group.addButton(button)
            self._amount_group.setId(button, int(value[:-1]))
            self._amount_group_layout.addWidget(button)
        self._amount_group.buttonClicked.connect(self._set_amount_from_button)

        self._price_label = QLabel("Price:")
        self._price_box = DecimalSpinBox()
        self._price_box.setDecimals(8)
        self._price_refresh_btn = QPushButton()
        self._close_btn = QPushButton("Close Trade")
        self._close_btn.setProperty("class", "gray")
        self._close_btn.setMinimumSize(50, 30)

        self._setup_widgets()
        self._setup_layout()

    def _setup_widgets(self) -> None:
        """Configure widgets."""
        self.type_group.addButton(self.limit_button)
        self.type_group.addButton(self.market_button)
        self.type_group_layout.addWidget(self.market_button)
        self.type_group_layout.addWidget(self.limit_button)
        self.type_group.buttonClicked.connect(self._type_change)
        self.market_button.click()

        self._amount_box.setDecimals(8)
        self._amount_box.setRange(Decimal(1), self._position["position_size_stable"])
        self._amount_box.decimalValueChanged.connect(self._update_amount_buttons)
        self._amount_box.setAlignment(Qt.AlignmentFlag.AlignRight)
        self._amount_group.button(100).setChecked(True)
        self._amount_box.setValue(self._position["position_size_stable"])
        self._price_box.setRange(Decimal(0), Decimal(100_000_000))
        self._price_box.setAlignment(Qt.AlignmentFlag.AlignRight)

        self._price_refresh_btn.setProperty("class", "borderless")
        self._price_refresh_btn.setIcon(
            QIcon(":/icons/focus_icon"),
        )
        self._price_refresh_btn.setFixedWidth(25)
        self._price_refresh_btn.clicked.connect(self.set_price_clicked.emit)

        self._close_btn.clicked.connect(self._reduce_position)

    def _setup_layout(self) -> None:
        """Configure layout."""
        self._main_layout.addLayout(self.type_group_layout, 0, 0, 1, 3)
        self._main_layout.addWidget(self._amount_label, 1, 0)
        self._main_layout.addWidget(self._amount_box, 1, 1, 1, 2)
        self._main_layout.addLayout(self._amount_group_layout, 2, 0, 1, 3)
        self._main_layout.addWidget(self._price_label, 3, 0)
        self._main_layout.addWidget(self._price_box, 3, 1)
        self._main_layout.addWidget(self._price_refresh_btn, 3, 2)
        self._main_layout.addWidget(self._close_btn, 4, 0, 1, 3)

        self._default_widget.setLayout(self._main_layout)
        self.setDefaultWidget(self._default_widget)

    def _type_change(self, button: QRadioButton) -> None:
        """Handle type change."""
        if button == self.limit_button:
            self._price_box.setEnabled(True)
            self._price_refresh_btn.setEnabled(True)
            self.set_price_clicked.emit()
        else:
            self._price_box.setEnabled(False)
            self._price_refresh_btn.setEnabled(False)
            self._price_box.setValue(Decimal(0))

    def _set_amount_spinbox(self, amount: Decimal) -> None:
        """Set amount on spinbox."""
        self._amount_box.blockSignals(True)
        self._amount_box.setValue(amount)
        self._amount_box.blockSignals(False)

    def _update_amount_buttons(self, amount: Decimal) -> None:
        """Update amount buttons if value on spinbox matches."""
        if amount == self._position["position_size_stable"] * Decimal(0.25):
            self._amount_group.button(25).setChecked(True)
        elif amount == self._position["position_size_stable"] * Decimal(0.5):
            self._amount_group.button(50).setChecked(True)
        elif amount == self._position["position_size_stable"] * Decimal(0.75):
            self._amount_group.button(75).setChecked(True)
        elif amount == self._position["position_size_stable"]:
            self._amount_group.button(100).setChecked(True)
        else:
            button = self._amount_group.checkedButton()
            if button:
                self._amount_group.setExclusive(False)
                button.setChecked(False)
                self._amount_group.setExclusive(True)

    def _set_amount_from_button(self, button: QRadioButton) -> None:
        """Set amount from button."""
        self._set_amount_spinbox(
            Decimal(self._amount_group.id(button) / 100) * self._position["position_size_stable"],
        )

    def _reduce_position(self) -> None:
        """Reduce position."""
        reduce_args: dict[str, Any] = {}
        reduce_args["pair"] = self._position["pair"]
        reduce_args["size"] = Decimal(self._amount_box.value())
        reduce_args["trade_direction"] = self._position["trade_direction"]
        if self.limit_button.isChecked():
            reduce_args["trade_type"] = PerpsTradeType.LIMIT
            reduce_args["execution_price"] = Decimal(self._price_box.value())
        else:
            reduce_args["trade_type"] = PerpsTradeType.MARKET
            reduce_args["execution_price"] = None

        self.reduce_clicked.emit(reduce_args)

    def set_price_limit(self, price: Decimal) -> None:
        """Set price for limit trade."""
        self._price_box.setValue(price)



"""Foxify exchange fetcher."""

from __future__ import annotations

import asyncio
from collections import defaultdict
from decimal import Decimal
import json
import logging
from pathlib import Path
import time
from typing import TYPE_CHECKING, Optional, Self

from httpx import AsyncClient, ConnectError, HTTPStatusError, ReadTimeout
import orjson
import pandas
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from websockets.client import WebSocketClientProtocol, connect
from websockets.exceptions import (
    ConnectionClosedError,
    InvalidStatus,
    InvalidStatusCode,
)

from plutus_terminal.core.exchange.base import ExchangeFetcher
from plutus_terminal.core.exchange.foxify import utils as foxify_utils
from plutus_terminal.core.exchange.types import OrderData, PerpsTradeType
from plutus_terminal.core.exchange.web3.cycle_provider import build_cycle_provider
from plutus_terminal.core.types_ import (
    PerpsPosition,
    PerpsTradeDirection,
    PriceData,
    PriceHistory,
)
from plutus_terminal.log_utils import log_retry

if TYPE_CHECKING:
    from eth_typing import ChecksumAddress

    from plutus_terminal.core.exchange.base import ExchangeFetcherMessageBus

LOGGER = logging.getLogger(__name__)


class FoxifyFetcher(ExchangeFetcher):
    """Fetches market information for Foxify Exchange."""

    def __init__(
        self,
        pair_map: dict[str, str],
        web3_address: ChecksumAddress,
        message_bus: ExchangeFetcherMessageBus,
    ) -> None:
        """Initialize shared attributes.

        Args:
            pair_map (dict[str, dict[str, str]]): Dict with adress and pair.
            web3_address (ChecksumAddress): Web3 account address.
            message_bus (ExchangeFetcherMessageBus): Message bus.
        """
        LOGGER.info("Initialize FoxifyFetcher")
        self.aclient = AsyncClient()
        self.pair_map = pair_map
        self.web3_address = web3_address
        self.web3_provider = build_cycle_provider("Arbitrum One Fetcher")
        self.vault_contract = foxify_utils.build_vault_contract(self.web3_provider)
        self.order_book_contract = foxify_utils.build_order_book_contract(
            self.web3_provider,
        )
        self.stable_contract = foxify_utils.build_stable_contract(self.web3_provider)

        with Path.open(Path(__file__).parent.parent.joinpath("web3/pyth_feed_id.json")) as f:
            self.pyth_pair_id: dict = json.load(f)
        self.pyth_id_pair = {value: key for key, value in self.pyth_pair_id.items()}
        self._message_bus = message_bus
        self._socket: Optional[WebSocketClientProtocol] = None  # type: ignore
        self.connection_count: dict = defaultdict(int)
        self._cached_prices: dict[str, PriceData] = {}
        self._cached_stable_balance: Decimal = Decimal(0)
        self._cached_positions: list[PerpsPosition] = []
        self._cached_funding_rates: dict[str, dict[bool, int]] = {}
        self._synced = True
        self._goldsky_url = "https://api.goldsky.com/api/public/project_cllqn805a7tva38uh4n5r9ap6/subgraphs/palmswap-synthetic-stats/arbitrum_mainnet/gn"
        self.async_stop_event = asyncio.Event()

    @classmethod
    async def create(
        cls,
        pair_map: dict[str, str],
        web3_address: ChecksumAddress,
        message_bus: ExchangeFetcherMessageBus,
    ) -> Self:
        """Create class instance and init_async.

        Args:
            pair_map (dict[str, str]): Dict with adress and pair.
            web3_address (ChecksumAddress): Web3 account address.
            message_bus (ExchangeFetcherMessageBus): Message bus.

        Returns:
            FoxifyFetcher: Instance of FoxifyFetcher.
        """
        fetcher = cls(pair_map, web3_address, message_bus)
        await fetcher.init_async()
        return fetcher

    @retry(
        wait=wait_exponential(multiplier=1, min=0.15, max=1),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def init_async(self) -> None:
        """Initialize async shared attributes."""
        self.time_lock_contract = await foxify_utils.build_time_lock_contract(self.web3_provider)
        (
            _,
            self._vault_price_precision,
            self._funding_rate_precision,
            self._basis_points_divisor,
            self._margin_fee_basis_points,
            self._liquidation_fee,
        ) = await asyncio.gather(
            self._populate_funding_rates(),
            self.vault_contract.functions.PRICE_PRECISION().call(),
            self.vault_contract.functions.FUNDING_RATE_PRECISION().call(),
            self.vault_contract.functions.BASIS_POINTS_DIVISOR().call(),
            self.time_lock_contract.functions.marginFeeBasisPoints().call(),
            self.vault_contract.functions.liquidationFeeUsd().call(),
        )

    @retry(
        retry=(
            retry_if_exception_type(HTTPStatusError)
            | retry_if_exception_type(ReadTimeout)
            | retry_if_exception_type(ConnectError)
        ),
        reraise=True,
        wait=wait_exponential(multiplier=1, min=0.15, max=1),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def fetch_price_history(
        self,
        pair: str,
        from_timestamp: int,
        resolution: str,
    ) -> PriceHistory:
        """Fetch the price history on PYTH Network.

        Args:
            pair (str): Pair to fetch price for.
            from_timestamp (int): Timestamp for the leftmost bar.
            resolution (str): Bar resolution in seconds.

        Returns:
            PriceHistory: Dict with ohlcv.

        Raises:
            HTTPStatusError: If response is not successful. After tries.
        """
        request_url = "https://benchmarks.pyth.network/v1/shims/tradingview/history"
        now_timestamp = int(time.time())
        request_params = {
            "symbol": pair,
            "resolution": resolution,
            "from": from_timestamp,
            "to": now_timestamp,
        }

        response = await self.aclient.get(
            request_url,
            params=request_params,  # type: ignore
        )
        response.raise_for_status()

        data = response.json()

        price_history: PriceHistory = {
            "date": [pandas.Timestamp(t, unit="s") for t in data["t"]],
            "open": data["o"],
            "high": data["h"],
            "low": data["l"],
            "close": data["c"],
        }
        LOGGER.debug(
            "Fetched price history for %s from %s",
            pair,
            from_timestamp,
        )
        return price_history

    @retry(
        retry=(
            retry_if_exception_type(ConnectionClosedError)
            | retry_if_exception_type(InvalidStatus)
            | retry_if_exception_type(asyncio.TimeoutError)
            | retry_if_exception_type(InvalidStatusCode)
            | retry_if_exception_type(ConnectionAbortedError)
        ),
        wait=wait_exponential(multiplier=1, min=1, max=5),
        reraise=True,
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def websocket_connect(self) -> WebSocketClientProtocol:
        """Connect to websocket to fetch prices.

        Returns:
            WebSocketClientProtocol: Websocket connection.
        """
        if self._socket is None or self._socket.closed:
            self._socket = await connect(
                "wss://hermes.pyth.network/ws",
                ping_interval=10,
                ping_timeout=10,
            )
            LOGGER.info("Connected to hermes.pyth.network for Foxify")
            self._message_bus.price_synced.emit(True)
        return self._socket

    @retry(
        wait=wait_exponential(multiplier=1, min=0.4, max=5),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def _ensure_websocket_connection(self) -> None:
        """Ensure websocket is connected."""
        if self._socket is None or self._socket.closed:
            LOGGER.warning(
                "Websocket disconnected. Attempting to reconnect websocket...",
            )

            self._message_bus.price_synced.emit(False)
            await self.websocket_connect()
            await self.resubscribe_on_going_connections()

    async def subscribe_to_price(self, pair: str, force: bool = False) -> None:
        """Subscribe to receive price updates for the given pair.

        Args:
            pair (str): Pair name e.g Crypto.BTC/USD.
            force (bool, optional): Force subscribe. Defaults to False.
        """
        await self._ensure_websocket_connection()

        self.connection_count[pair] += 1
        # Don't subscribe again if already connected.
        if self.connection_count[pair] > 1 and not force:
            return
        await self._socket.send(  # type: ignore
            json.dumps(
                {
                    "ids": [self.pyth_pair_id[pair]],
                    "type": "subscribe",
                },
            ),
        )
        LOGGER.info("Subscribed to receive price updates of: %s", pair)

    async def resubscribe_on_going_connections(self) -> None:
        """Resubscribe on going connections."""
        LOGGER.warning("Re-subscribing to on going connection...")
        for pair, count in self.connection_count.items():
            if count >= 1:
                await self.subscribe_to_price(pair, force=True)
                # Reduce 1 since this connection should already be present
                self.connection_count[pair] -= 1

    async def unsubscribe_to_price(self, pair: str, force: bool = False) -> None:
        """Unsubscribe to receive price updates for the given pair.

        Args:
            pair (str): Pair name e.g Crypto.BTC/USD.
            force (bool, optional): Force unsubscribe. Defaults to False.
        """
        await self._ensure_websocket_connection()

        self.connection_count[pair] -= 1
        if self.connection_count[pair] > 0 and not force:
            return

        await self._socket.send(  # type: ignore
            json.dumps(
                {
                    "ids": [self.pyth_pair_id[pair]],
                    "type": "unsubscribe",
                },
            ),
        )
        self._cached_prices.pop(pair)

        self.connection_count[pair] = 0
        LOGGER.info("Unsubscribed to receive price updates of: %s", pair)

    async def receive_subscribed_prices(self) -> None:
        """Receive live data of subscribed prices from exchange."""
        await self._ensure_websocket_connection()
        while not self.async_stop_event.is_set():
            try:
                async for message in self._socket:  # type: ignore
                    data = orjson.loads(message)
                    if data.get("type", "") == "price_update":
                        pair = self.pyth_id_pair[f'0x{data["price_feed"]["id"]}']
                        self._cached_prices[pair] = PriceData(
                            {
                                "price": Decimal(
                                    int(data["price_feed"]["price"]["price"])
                                    * 10 ** int(data["price_feed"]["price"]["expo"]),
                                ),
                                "date": pandas.Timestamp(
                                    data["price_feed"]["price"]["publish_time"],
                                    unit="s",
                                ),
                            },
                        )
                        self._message_bus.subscribed_prices_signal.emit(
                            self._cached_prices,
                        )
            except (ConnectionClosedError, ConnectionAbortedError):
                LOGGER.exception("WebSocket connection closed unexpectely")
                await self._ensure_websocket_connection()
            except asyncio.CancelledError:
                LOGGER.debug("Receive subscribed prices cancelled.")
                raise
            except Exception:
                LOGGER.exception("Unexpected error while receiving prices")

    @retry(
        retry=(
            retry_if_exception_type(HTTPStatusError)
            | retry_if_exception_type(ReadTimeout)
            | retry_if_exception_type(ConnectError)
        ),
        reraise=True,
        wait=wait_exponential(multiplier=1, min=0.15, max=1),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def fetch_price_at_time(self, pair: str, timestamp: int) -> PriceData:
        """Fetch price at current time.

        Args:
            pair (str): Pair name e.g Crypto.BTC/USD.
            timestamp (int): Unix timestamp in seconds.

        Returns:
            PriceData: Price data for fetched pair.
        """
        LOGGER.debug("Fetching %s price at time: %s", pair, timestamp)
        request_url = f"https://hermes.pyth.network/v2/updates/price/{int(timestamp)}"
        request_params = {
            "ids[]": self.pyth_pair_id[pair],
        }

        response = await self.aclient.get(request_url, params=request_params)
        response.raise_for_status()

        data = response.json()
        return PriceData(
            {
                "price": Decimal(
                    int(data["parsed"][0]["price"]["price"])
                    * 10 ** int(data["parsed"][0]["price"]["expo"]),
                ),
                "date": data["parsed"][0]["price"]["publish_time"],
            },
        )

    async def watch_all_positions(self) -> None:
        """Watch all open positions for all available pairs."""
        while not self.async_stop_event.is_set():
            try:
                all_positions = await self.fetch_all_positions()
                self._message_bus.positions_signal.emit(all_positions)
            except (HTTPStatusError, ReadTimeout, ConnectError):
                LOGGER.exception("Unexpected error while fetching all positions")
                continue
            except asyncio.CancelledError:
                LOGGER.debug("Watch all positions cancelled.")
                raise
            finally:
                await asyncio.sleep(1)

    async def refresh_all_positions(self) -> None:
        """Refresh all open positions for all available pairs.

        This is used to force a refresh of all positions on the UI.
        """
        try:
            all_positions = await self.fetch_all_positions()
            self._message_bus.positions_signal.emit(all_positions)
        except (HTTPStatusError, ReadTimeout, ConnectError):
            LOGGER.exception("Unexpected error while fetching all positions")

    @retry(
        wait=wait_exponential(multiplier=1, min=0.15, max=5),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def fetch_all_positions(self) -> list[PerpsPosition]:
        """Fetch all open positions for all available pairs."""
        query = """
        query activePositionsByAccount($account: String!) {
          activePositions(
            where: {account: $account, collateral_not: "0"}
          ) {
            id
            averagePrice
            isLong
            size
            indexToken
            entryFundingRate
            collateral
            account
          }
        }
        """

        variables = {"account": self.web3_address.lower()}
        response = await self.aclient.post(
            self._goldsky_url,
            json={"query": query, "variables": variables},
        )
        response.raise_for_status()
        data = response.json()
        positions_data = data["data"]["activePositions"]

        self._cached_positions.clear()
        for data in positions_data:
            pair = self.pair_map[data["indexToken"]]
            position_size = Decimal(data["size"]) / self._vault_price_precision
            leverage = position_size / (
                Decimal(data["collateral"]) / Decimal(self._vault_price_precision)
            )
            extra = {}
            extra["entry_funding_rate"] = data["entryFundingRate"]
            extra["index_token"] = data["indexToken"]
            collateral = position_size / leverage

            position = PerpsPosition(
                {
                    "pair": pair,
                    "id": data["id"],
                    "position_size_stable": position_size,
                    "collateral_stable": collateral,
                    "open_price": Decimal(data["averagePrice"]) / self._vault_price_precision,
                    "trade_direction": (
                        PerpsTradeDirection.LONG if data["isLong"] else PerpsTradeDirection.SHORT
                    ),
                    "leverage": round(leverage, 2),
                    "extra": extra,
                    "liquidation_price": Decimal(0),
                },
            )
            position["liquidation_price"] = self.calculate_liquidation_price(position)
            self._cached_positions.append(position)
        return self._cached_positions

    async def watch_all_orders(self) -> None:
        """Watch all open orders for all available pairs."""
        while not self.async_stop_event.is_set():
            try:
                all_orders = await self.fetch_all_orders()
                self._message_bus.orders_signal.emit(all_orders)
            except (HTTPStatusError, ReadTimeout, ConnectError):
                LOGGER.exception("Unexpected error while fetching all positions")
                continue
            except asyncio.CancelledError:
                LOGGER.debug("Watch all order cancelled.")
                raise
            finally:
                await asyncio.sleep(1)

    async def refresh_all_orders(self) -> None:
        """Refresh all open orders for all available pairs.

        This is used to force a refresh of all orders on the UI.
        """
        try:
            all_orders = await self.fetch_all_orders()
            self._message_bus.orders_signal.emit(all_orders)
        except (HTTPStatusError, ReadTimeout, ConnectError):
            LOGGER.exception("Unexpected error while fetching all positions")

    @retry(
        wait=wait_exponential(multiplier=1, min=0.15, max=5),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def fetch_all_orders(self) -> list[OrderData]:
        """Fetch all open orders."""
        query = """
        query WatchOrders($account: String!) {
          orders(

            orderBy: createdTimestamp
            orderDirection: desc

            where: {account: $account, status: open}
          ) {
            id
            index
            indexToken
            size
            type
            status
            createdTimestamp
          }
        }
        """
        variables = {"account": self.web3_address.lower()}
        response = await self.aclient.post(
            self._goldsky_url,
            json={"query": query, "variables": variables},
        )
        response.raise_for_status()
        data = response.json()
        order_data = data["data"]["orders"]
        all_orders = []
        for order in order_data:
            func = self.order_book_contract.get_function_by_name(
                f"get{order['type'].capitalize()}Order",
            )
            chain_order = await func(
                self.web3_address,
                int(order["index"]),
            ).call()

            reduce_only = order["type"] == "decrease"
            is_long = chain_order[4]
            trigger_price = Decimal(chain_order[5])
            trigger_above_threshold = chain_order[6]
            extras = {
                "is_long": is_long,
                "trigger_above_threshold": trigger_above_threshold,
                "index": order["index"],
            }

            order_type = PerpsTradeType.LIMIT
            if trigger_price > 0 and reduce_only:
                if (is_long and trigger_above_threshold) or (
                    not is_long and not trigger_above_threshold
                ):
                    order_type = PerpsTradeType.TRIGGER_TP
                else:
                    order_type = PerpsTradeType.TRIGGER_SL

            all_orders.append(
                OrderData(
                    {
                        "id": order["id"],
                        "pair": self.pair_map[order["indexToken"]],
                        "trigger_price": trigger_price / self._vault_price_precision,
                        "size_stable": Decimal(order["size"]) / self._vault_price_precision,
                        "trade_direction": (
                            PerpsTradeDirection.LONG if is_long else PerpsTradeDirection.SHORT
                        ),
                        "order_type": order_type,
                        "reduce_only": reduce_only,
                        "extra": extras,
                    },
                ),
            )
        return all_orders

    def get_position_associated_with_order(self, order: OrderData) -> Optional[PerpsPosition]:
        """Get position associated with given order.

        Args:
            order (OrderData): Order to get position for.

        Returns:
            Optional[PerpsPosition]: Position associated with given order.
        """
        for position in self._cached_positions:
            if (
                order["pair"] == position["pair"]
                and order["trade_direction"] == position["trade_direction"]
            ):
                return position
        return None

    @retry(
        reraise=True,
        stop=stop_after_attempt(3),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def fetch_current_price(self, pair: str) -> PriceData:
        """Fetch current price of given pair."""
        return await self.fetch_price_at_time(pair=pair, timestamp=int(time.time()))

    async def watch_stable_balance(self) -> None:
        """Watch stable balance."""
        while not self.async_stop_event.is_set():
            try:
                self._cached_stable_balance = await self.fetch_stable_balance()
                self._message_bus.balance_signal.emit(self._cached_stable_balance)
            except (HTTPStatusError, ReadTimeout, ConnectError):
                LOGGER.exception("Unexpected error while fetching stable balance.")
                continue
            except asyncio.CancelledError:
                LOGGER.debug("Watch all balance cancelled.")
                raise
            finally:
                await asyncio.sleep(2)

    @retry(
        wait=wait_exponential(multiplier=1, min=0.15, max=5),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def fetch_stable_balance(self) -> Decimal:
        """Fetch stable balance.

        Returns:
            Decimal: Balance in USD Stable Format.
        """
        balance = await self.stable_contract.functions.balanceOf(self.web3_address).call()

        return Decimal(balance) / 10**foxify_utils.USDC_DECIMAL_PLACES

    def calculate_margin_fee(self, position_size: Decimal) -> Decimal:
        """Calulate fee for a given position.

        Args:
            position_size (Decimal): Position size to get fee for.

        Returns:
            Decimal: Fee amount in USD Stable Format.
        """
        return (
            position_size
            * (
                self._basis_points_divisor
                - (self._basis_points_divisor - self._margin_fee_basis_points)
            )
        ) / (self._basis_points_divisor)

    @retry(
        wait=wait_exponential(multiplier=1, min=0.15, max=5),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def fetch_positions_funding_rates(self) -> None:
        """Fetch funding rate of cached positions.."""
        while True:
            await self._fetch_positions_funding_rates()
            await asyncio.sleep(5)

    async def _fetch_positions_funding_rates(self) -> None:
        """Fetch funding rate of cached positions.."""
        for position in self._cached_positions:
            position_extra = position.get("extra", None)
            if position_extra is None:
                LOGGER.error("Position extra data not found")
                continue

            self._cached_funding_rates[position["pair"]][position["trade_direction"].value] = int(
                await self.vault_contract.functions.cumulativeFundingRates(
                    self.web3_provider.to_checksum_address(
                        position_extra["index_token"],
                    ),
                    position["trade_direction"].value,
                ).call(),
            )

    async def _populate_funding_rates(self) -> None:
        """Fetch funding rate of cached positions."""
        tasks = []
        for direction in [
            PerpsTradeDirection.LONG.value,
            PerpsTradeDirection.SHORT.value,
        ]:
            for pair_address, pair in self.pair_map.items():
                self._cached_funding_rates.setdefault(pair, {})
                tasks.append(self._fetch_and_cache_funding_rate(pair, pair_address, direction))

        await asyncio.gather(*tasks)

    async def _fetch_and_cache_funding_rate(
        self,
        pair: str,
        pair_address: str,
        direction: bool,
    ) -> None:
        """Fetch and cache funding rate for a specific pair and direction.

        Args:
            pair (str): Pair name e.g Crypto.BTC/USD.
            pair_address (str): Pair address.
            direction (bool): Direction to fetch funding rate for.
        """
        rate = int(
            await self.vault_contract.functions.cumulativeFundingRates(
                self.web3_provider.to_checksum_address(pair_address),
                direction,
            ).call(),
        )
        self._cached_funding_rates[pair][direction] = rate

    def fetch_funding_fee(self, perps_position: PerpsPosition) -> Decimal:
        """Get funding fee for a given position.

        Args:
            perps_position (PerpsPosition): Position to get funding fee for.

        Returns:
            Decimal: Funding fee in USD Stable Format.
        """
        position_extra = perps_position.get("extra", None)
        # If no position extra data, consider no funding fee
        if position_extra is None:
            return Decimal(0)

        cumulative_funding_rate = self._cached_funding_rates[perps_position["pair"]][
            perps_position["trade_direction"].value
        ]

        funding_fee = cumulative_funding_rate - int(
            position_extra["entry_funding_rate"],
        )

        size = perps_position["position_size_stable"]
        return (size * funding_fee) / self._funding_rate_precision

    def calculate_liquidation_price(self, perps_position: PerpsPosition) -> Decimal:
        """Calculate liquidation price for a given position.

        This formula is not perfect and should be improved, values don't match
        the ones given on the front end, but are close.

        Args:
            perps_position (PerpsPosition): Position to get liquidation price for.

        Returns:
            Decimal: Liquidation price in USD Stable Format.
        """
        collateral = perps_position["collateral_stable"]
        funding_fee = self.fetch_funding_fee(perps_position)
        margin_fee = self.calculate_margin_fee(perps_position["position_size_stable"])
        liquidation_fee = Decimal(self._liquidation_fee / self._vault_price_precision)
        size = perps_position["position_size_stable"]
        trade_direction = perps_position["trade_direction"]
        total_fees = funding_fee + margin_fee + liquidation_fee

        open_price = perps_position["open_price"]

        liquidation_price_for_fees = self._calculate_liquidation_price_from_delta(
            total_fees,
            size,
            collateral,
            open_price,
            trade_direction,
        )

        liquidation_price_for_max_leverage = self._calculate_liquidation_price_from_delta(
            size * self._basis_points_divisor / 1000000,
            size,
            collateral,
            open_price,
            trade_direction,
        )

        if trade_direction == PerpsTradeDirection.LONG:
            return max(liquidation_price_for_fees, liquidation_price_for_max_leverage)
        return min(liquidation_price_for_fees, liquidation_price_for_max_leverage)

    def _calculate_liquidation_price_from_delta(
        self,
        total_fees: Decimal,
        size: Decimal,
        collateral: Decimal,
        open_price: Decimal,
        trade_direction: PerpsTradeDirection,
    ) -> Decimal:
        """Calculate liquidation price from delta.

        Args:
            total_fees (Decimal): Total fees (collateral + position fee + liquidation fee).
            size (Decimal): Position size.
            collateral (Decimal): Collateral.
            open_price (Decimal): Open price.
            trade_direction (PerpsTradeDirection): Trade direction.

        Returns:
            Decimal: Liquidation price in USD Stable Format.
        """
        if total_fees > collateral:
            if trade_direction == PerpsTradeDirection.LONG:
                return open_price - ((total_fees - collateral) * open_price / size)
            return open_price + ((total_fees - collateral) * open_price / size)

        if trade_direction == PerpsTradeDirection.LONG:
            return open_price - ((collateral - total_fees) * open_price / size)
        return open_price + ((collateral - total_fees) * open_price / size)

    def calculate_pnl_percent_before_fees(
        self,
        perps_position: PerpsPosition,
        current_price: Optional[Decimal],
    ) -> Decimal:
        """Calculate pnl percent for a given position.

        Args:
            perps_position (PerpsPosition): Position to get pnl for.
            current_price (Optional[Decimal]): Current price of the pair.

        Returns:
            Decimal: PNL in USD Stable Format.
        """
        open_price = perps_position["open_price"]
        trade_direction = perps_position["trade_direction"]
        leverage = perps_position["leverage"]
        trade_direction_multiplier = 1 if trade_direction == PerpsTradeDirection.LONG else -1
        if current_price is None:
            current_price = self._cached_prices[perps_position["pair"]]["price"]
        return ((current_price / open_price) - 1) * 100 * leverage * trade_direction_multiplier

    async def stop_async(self) -> None:
        """Stop infinite loops and close connections."""
        self.async_stop_event.set()
        if self._socket is not None:
            await self._socket.close()
        await self.aclient.aclose()



"""Bitget Exchange."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional

from ccxt.base.errors import ExchangeError
from ccxt.base.exchange import Decimal
import ccxt.pro
import keyring
import orjson
from qasync import asyncSlot

from plutus_terminal.core.config import CONFIG
from plutus_terminal.core.exceptions import TransactionFailedError
from plutus_terminal.core.exchange.base import ExchangeBase
from plutus_terminal.core.exchange.ccxt import utils as ccxt_utils
from plutus_terminal.core.exchange.ccxt.fetcher import CCXTFetcher
from plutus_terminal.core.exchange.ccxt.trader import CCXTTrader
from plutus_terminal.core.types_ import (
    ExchangeType,
    NewAccountInfo,
    PerpsTradeDirection,
    PerpsTradeType,
)
from plutus_terminal.ui.widgets.toast import Toast, ToastType

if TYPE_CHECKING:
    from typing import Self

    from plutus_terminal.core.exchange.base import ExchangeFetcherMessageBus

LOGGER = logging.getLogger(__name__)


class BitgetExchange(ExchangeBase):
    """Class to interact with Bitget exchange."""

    def __init__(self, fetcher_bus: ExchangeFetcherMessageBus) -> None:
        """Initialize shared variables."""
        super().__init__(fetcher_bus)
        keyring_account = CONFIG.current_keyring_account
        keyring_password = keyring.get_password(
            "plutus-terminal",
            str(keyring_account.username),
        )
        if keyring_password is None:
            msg = "Keyring password not found"
            raise Exception(msg)

        api_key, secret, password = orjson.loads(keyring_password)
        self.bidget_ccxt = ccxt.pro.bitget(
            {
                "apiKey": api_key,
                "secret": secret,
                "password": password,
                "enableRateLimit": True,
            },
        )

        self._pair_prefix = ""
        self._pair_separator = "/"
        self._quote_symbol = "USDT"
        self._pair_suffix = ":USDT"

        self._fetcher = CCXTFetcher(fetcher_bus, self.bidget_ccxt)
        self._trader = CCXTTrader(self.bidget_ccxt)

    @classmethod
    async def create(cls, fetcher_bus: ExchangeFetcherMessageBus) -> Self:
        """Create class instance and init_async.

        Args:
            fetcher_bus (ExchangeFetcherMessageBus): ExchangeFetcherMessageBus.

        Returns:
            CCXTExchange: Instance of FoxifyExchange.
        """
        instance = cls(fetcher_bus)
        await instance.init_async()
        return instance

    async def init_async(self) -> None:
        """Initialize async shared attributes."""
        await self.bidget_ccxt.load_markets()

    @property
    def name(self) -> str:
        """Returns: Exchange name."""
        return "bitget"

    @property
    def trader(self) -> CCXTTrader:
        """Returns: Exchange Trader."""
        return self._trader

    @property
    def fetcher(self) -> CCXTFetcher:
        """Returns: Exchange Trader."""
        return self._fetcher

    @property
    def available_pairs(self) -> set:
        """Returns set with all available pairs."""
        all_symbols = self.bidget_ccxt.symbols
        if not all_symbols:
            return set()
        return {
            symbol
            for symbol in all_symbols
            if symbol[: -len(self._pair_suffix)].endswith(self._quote_symbol)
        }

    @property
    def pair_prefix(self) -> str:
        """Returns pair prefix if any."""
        return self._pair_prefix

    @property
    def pair_separator(self) -> str:
        """Returns the pair separator if any."""
        return self._pair_separator

    @property
    def quote_symbol(self) -> str:
        """Return quote symbol."""
        return self._quote_symbol

    @property
    def pair_suffix(self) -> str:
        """Returns pair suffix."""
        return self._pair_suffix

    @property
    def cached_prices(self) -> dict:
        """Return all prices cache."""
        return self.fetcher._cached_prices  # noqa: SLF001

    @asyncSlot()
    async def set_all_leverage(self, leverage: int) -> None:
        """Set leverage for all positions.

        Args:
            leverage (int): Leverage to set.
        """
        toast_id = Toast.show_message(
            f"Setting leverage to all pairs: {leverage}x",
            type_=ToastType.WARNING,
        )
        for x, pair in enumerate(self.available_pairs):
            Toast.update_message(
                toast_id,
                f"Setting leverage to {leverage}x {x}/{len(self.available_pairs)}: {pair}",
                type_=ToastType.WARNING,
            )
            try:
                await self.bidget_ccxt.set_leverage(leverage, pair)
                LOGGER.debug("Set leverage for %s to %s", pair, leverage)
            except ExchangeError:
                # If error fetch max leverage and set it
                Toast.update_message(
                    toast_id,
                    f"Error setting leverage for pair: {pair}",
                    type_=ToastType.ERROR,
                )
                leverage_tiers = await self.bidget_ccxt.fetch_market_leverage_tiers(
                    pair,
                )
                max_leverage = int(leverage_tiers[0]["maxLeverage"])  # type: ignore

                Toast.update_message(
                    toast_id,
                    f"Setting leverage to {max_leverage}x {x}/{len(self.available_pairs)}: {pair}",
                    type_=ToastType.WARNING,
                )
                await self.bidget_ccxt.set_leverage(max_leverage, pair)
                LOGGER.debug("Set leverage for %s to %s", pair, max_leverage)

        CONFIG.leverage = leverage

        Toast.update_message(
            toast_id,
            f"Leverage set to all pairs: {leverage}x",
            type_=ToastType.SUCCESS,
        )

    @asyncSlot()
    async def set_leverage(self, coin: str, leverage: int) -> None:
        """Set leverage for pair.

        Args:
            coin (str): Coin to set leverage for.
            leverage (int): Leverage to set.
        """
        pair = self.format_pair_from_coin(coin)
        toast_id = Toast.show_message(
            f"Setting leverage from {pair} to {leverage}x",
            type_=ToastType.WARNING,
        )
        try:
            await self.bidget_ccxt.set_leverage(leverage, pair)
            LOGGER.debug("Set leverage from %s to %s", pair, leverage)
        except ExchangeError:
            # If error fetch max leverage and set it
            Toast.update_message(
                toast_id,
                f"Error setting leverage for: {pair}",
                type_=ToastType.ERROR,
            )
            leverage_tiers = await self.bidget_ccxt.fetch_market_leverage_tiers(pair)
            max_leverage = int(leverage_tiers[0]["maxLeverage"])  # type: ignore

            Toast.update_message(
                toast_id,
                f"Setting leverage of {pair} to {max_leverage}x",
                type_=ToastType.WARNING,
            )
            await self.bidget_ccxt.set_leverage(max_leverage, pair)
            LOGGER.debug("Set leverage for %s to %s", pair, max_leverage)

        Toast.update_message(
            toast_id,
            f"Leverage of {pair} set to: {leverage}x",
            type_=ToastType.SUCCESS,
        )
        CONFIG.leverage = leverage

    @asyncSlot()
    async def create_order(
        self,
        pair: str,
        amount: Decimal,
        trade_direction: PerpsTradeDirection,
        trade_type: PerpsTradeType,
        execution_price: Optional[Decimal] = None,
    ) -> None:
        """Open position for trade.

        Args:
            pair (str): Pair to open trade for.
            amount (Decimal): Value in stable to open trade for, this will be multiplied
                by de configured leverage.
            trade_direction (TradeDirection): Trade direction.
            trade_type (TradeType): Trade type.
            execution_price (Optional[Decimal], optional): Execution price.
        """
        toast_id = Toast.show_message(
            f"Opening position for {pair}",
            type_=ToastType.WARNING,
        )

        # Get cached price if available otherwise fetch it.
        try:
            current_price = Decimal(self.cached_prices[pair]["price"])
        except KeyError:
            current_price_data = await self.fetcher.fetch_current_price(pair)
            current_price = Decimal(current_price_data["price"])

        if trade_direction == PerpsTradeDirection.LONG:
            if CONFIG.take_profit == 0:
                take_profit = current_price
            else:
                take_profit = current_price * (1 + Decimal(CONFIG.take_profit / 100))
            if CONFIG.stop_loss == 0:
                stop_loss = current_price
            else:
                stop_loss = current_price * (1 - Decimal(CONFIG.stop_loss / 100))
        else:
            if CONFIG.take_profit == 0:
                take_profit = current_price
            else:
                take_profit = current_price * (1 - Decimal(CONFIG.take_profit / 100))
            if CONFIG.stop_loss == 0:
                stop_loss = current_price
            else:
                stop_loss = current_price * (1 + Decimal(CONFIG.stop_loss / 100))

        # Convert amount with leverage to base asset
        amount = (amount * Decimal(CONFIG.leverage)) / current_price

        trading_args = ccxt_utils.TradingArgs(
            {
                "pair": pair,
                "trade_type": trade_type,
                "side": trade_direction,
                "price": current_price,
                "amount": amount,
                "take_profit": take_profit,
                "stop_loss": stop_loss,
            },
        )
        try:
            await self.trader.create_order(trading_args)
        except TransactionFailedError as error:
            Toast.update_message(
                toast_id,
                f"Failed to open position > {error}",
                type_=ToastType.ERROR,
            )
        Toast.update_message(
            toast_id,
            f"Position opened for {pair}",
            type_=ToastType.SUCCESS,
        )

    async def stop_async(self) -> None:
        """Stop all async tasks and cleanup for deletion."""
        await super().stop_async()
        await self.bidget_ccxt.close()

    @staticmethod
    def new_account_info() -> NewAccountInfo:
        """Provide info for new account creation."""
        return NewAccountInfo(
            referral_link="",
            secrets=["API Key", "Secret", "Passphrase"],
        )

    @staticmethod
    def exchange_type() -> ExchangeType:
        """Return exchange type."""
        return ExchangeType.CEX

    @staticmethod
    def validate_secrets(secrets: list[str]) -> tuple[bool, str]:
        """Validate secrets."""
        import ccxt

        try:
            exchange = ccxt.bitget(
                {
                    "apiKey": secrets[0],
                    "secret": secrets[1],
                    "password": secrets[2],
                    "enableRateLimit": True,
                },
            )
            exchange.fetch_balance()
        except Exception:  # noqa: BLE001
            return False, "Invalid API credentials"
        else:
            return True, "Valid API credentials."



"""TrandingChart Widget to visualize price data."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional

from lightweight_charts.widgets import QtChart
import pandas
from PySide6.QtCore import QEvent, QObject, Qt, Signal
from PySide6.QtGui import (
    QCloseEvent,
    QKeyEvent,
    QKeySequence,
    QMouseEvent,
    QPixmap,
    QShortcut,
    QShowEvent,
)
from PySide6.QtWidgets import (
    QApplication,
    QCompleter,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from plutus_terminal.ui import ui_utils
from plutus_terminal.ui.widgets.top_bar_widget import TopBar

if TYPE_CHECKING:
    from collections.abc import Callable

    from lightweight_charts import Chart
    from lightweight_charts.abstract import HorizontalLine

    from plutus_terminal.core.exchange.base import ExchangeBase
    from plutus_terminal.core.exchange.types import OrderData, PerpsPosition

LOGGER = logging.getLogger(__name__)


class TradingChart(QWidget):
    """Trading Chart Widget."""

    timeframe_signal = Signal(str)
    pair_changed = Signal(str)

    def __init__(
        self,
        available_pairs: set[str],
        format_simple_pair: Callable[[str], str],
        parent: Optional[QWidget] = None,
    ) -> None:
        """Initialize shared attributes."""
        super().__init__(parent=parent)
        self._available_pairs = available_pairs
        self._format_simple_pair = format_simple_pair

        self._main_layout = QVBoxLayout()

        self.top_bar = TopBar("Chart")
        self._price_label = QLabel("")

        self._main_chart = QtChart(toolbox=True)
        self._current_pair = ""
        self._position_lines: dict[int, HorizontalLine] = {}
        self._liquidation_lines: dict[int, HorizontalLine] = {}
        self._order_lines: dict[str, HorizontalLine] = {}

        self._config_widgets()
        self._config_chart()
        self._config_layout()
        self._config_shortcuts()
        self.setLayout(self._main_layout)

        self.setMinimumHeight(self.sizeHint().height() // 2)
        self.resize(self.sizeHint())

    def _config_widgets(self) -> None:
        """Configure widgets."""
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.top_bar.icon.setPixmap(QPixmap(":/icons/chart_icon"))
        self.top_bar.add_widget(self._price_label)
        self._price_label.setAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        self._price_label.setObjectName("title")

    def _config_chart(self) -> None:
        """Configure chart."""
        self._main_chart.precision(4)
        self._main_chart.layout(background_color="#131722")
        self._main_chart.legend(visible=True, font_size=14)
        self._main_chart.topbar.textbox("pair")
        self._main_chart.topbar.switcher(
            "timeframe",
            ("1min", "5min", "15min", "30min", "1hr", "4hr"),
            default="1min",
            func=self.on_timeframe_selection,
        )

    def _config_layout(self) -> None:
        """Configure layout."""
        self._main_layout.addWidget(self.top_bar)
        self._main_layout.addWidget(self._main_chart.get_webview())

    def _config_shortcuts(self) -> None:
        """Configure shortcuts."""
        self._shortcut = QShortcut(QKeySequence("\\"), self)
        self._shortcut.activated.connect(self.show_search_pair)

    @property
    def main_chart(self) -> QtChart:
        """Returns: Main QtChart widget."""
        return self._main_chart

    @property
    def current_pair(self) -> str:
        """Returns the pair being diplayed."""
        return self._current_pair

    @current_pair.setter
    def current_pair(self, pair: str) -> None:
        """Set the current pair being displayed."""
        self._current_pair = pair
        self.set_pair_text(pair)

    @property
    def current_timeframe(self) -> str:
        """Returns the current timeframe."""
        timeframe_value = self._main_chart.topbar["timeframe"].value
        if timeframe_value.endswith("min"):
            timeframe_value = timeframe_value[:-3]
        else:
            timeframe_value = timeframe_value[:-2]
            timeframe_value = int(timeframe_value) * 60

        return str(timeframe_value)

    def set_pair_text(self, pair: str) -> None:
        """Fill topbar text with pair."""
        pair_text = self._format_simple_pair(pair)
        self._main_chart.topbar["pair"].set(pair_text)  # type: ignore
        self.top_bar.title.setText(f"Chart | {pair_text}")

    def set_start_data(self, ohlcv: pandas.DataFrame, reset: bool = False) -> None:
        """Clean chart and fill with start data.

        Args:
            ohlcv (pandas.DataFrame): Open, high, low, close, volume data.
            reset (bool, optional): Reset chart for new symbol. Defaults to False.
        """
        self._main_chart.set(ohlcv)
        self._main_chart.price_scale()
        self._main_chart.fit()
        if reset:
            # TODO: remove draw_positions
            pass

    def update_chart_ohlcv(self, ohlcv: pandas.DataFrame) -> None:
        """Update the chart with the ohlcv data.

        Args:
            ohlcv (pandas.DataFrame): Open, high, low, close, volume data.
        """
        self._main_chart.update(ohlcv)  # type: ignore

    def update_chart_tick(self, data: dict) -> None:
        """Update the chart with the price tick.

        Args:
            data (dict): Data with all available prices.
        """
        try:
            price = data[self.current_pair]
        except KeyError:
            LOGGER.warning(
                "Price data for %s not available. Skipping update.",
                self.current_pair,
            )
            return
        tick = pandas.Series(price)
        # Convert decimal to float
        tick["price"] = float(tick["price"])
        self._main_chart.update_from_tick(tick)

        minimal_digits = ui_utils.get_minimal_digits(tick["price"], 4)
        self._price_label.setText(f"${tick['price']:,.{minimal_digits}f}")

    def draw_positions(self, all_positions: list[PerpsPosition]) -> None:
        """Draw positions lines on the chart.

        Args:
            all_positions (list[PerpsPosition]): List of current positions.
        """
        new_positions = {
            pos["id"]: pos for pos in all_positions if pos["pair"] == self.current_pair
        }

        # Update existing positions
        for pos_id, position in new_positions.items():
            if pos_id in self._position_lines:
                # Update position if open price has changed
                current_line = self._position_lines[pos_id]
                if float(position["open_price"]) != current_line.price:
                    current_line.update(float(position["open_price"]))

                # Update liquidation line if price has changed
                liquidation_price = position["liquidation_price"]
                current_liquidation_line = self._liquidation_lines[pos_id]
                if float(liquidation_price) != current_liquidation_line.price:
                    current_liquidation_line.update(float(liquidation_price))
            else:
                # Create new position line
                new_pos_line = self._main_chart.horizontal_line(
                    float(position["open_price"]),
                    width=1,
                    color="rgb(255, 80, 80)",
                    style="dotted",
                    text=f"Open {position['trade_direction'].name.capitalize()}",
                    axis_label_visible=False,
                )
                self._position_lines[pos_id] = new_pos_line

                # Create new liquidation line
                liquidation_price = position["liquidation_price"]
                new_liquidation_line = self._main_chart.horizontal_line(
                    float(liquidation_price),
                    width=1,
                    color="rgb(225, 110, 30)",
                    style="solid",
                    text=f"Liq. {position['trade_direction'].name.capitalize()}",
                    axis_label_visible=False,
                )
                self._liquidation_lines[pos_id] = new_liquidation_line

        # Delete old lines
        positions_to_delete = [
            pos_id for pos_id in self._position_lines if pos_id not in new_positions
        ]

        for pos_id in positions_to_delete:
            self._position_lines[pos_id].delete()
            del self._position_lines[pos_id]
            self._liquidation_lines[pos_id].delete()
            del self._liquidation_lines[pos_id]

    def draw_orders(self, all_orders: list[OrderData]) -> None:
        """Draw orders lines on the chart.

        Args:
            all_orders (list[OrderData]): List of current orders.
        """
        new_orders = {
            order["id"]: order for order in all_orders if order["pair"] == self.current_pair
        }

        for order_id, order in new_orders.items():
            if order_id in self._order_lines:
                # Update order if price has changed
                current_line = self._order_lines[order_id]
                if float(order["trigger_price"]) != current_line.price:
                    current_line.update(float(order["trigger_price"]))
            else:
                # Create new order line
                self._order_lines[order_id] = self._main_chart.horizontal_line(
                    float(order["trigger_price"]),
                    width=1,
                    color="rgb(255, 80, 80)",
                    style="dashed",
                    text=f"{order['order_type'].name.replace('_', ' ').capitalize()}",
                    axis_label_visible=False,
                )

        # Delete old lines
        order_to_delete = [order_id for order_id in self._order_lines if order_id not in new_orders]

        for order_id in order_to_delete:
            self._order_lines[order_id].delete()
            del self._order_lines[order_id]

    def on_timeframe_selection(self, chart: Chart) -> None:
        """Emit signal to change timeframe."""
        timeframe_value = chart.topbar["timeframe"].value
        if timeframe_value.endswith("min"):
            timeframe_value = timeframe_value[:-3]
        else:
            timeframe_value = timeframe_value[:-2]
            timeframe_value = int(timeframe_value) * 60

        self.timeframe_signal.emit(str(timeframe_value))

    def show_search_pair(self) -> None:
        """Show search pair modal."""
        modal_children = self.findChildren(SearchPairModal)
        if modal_children:
            modal_children[0].search_input.setFocus()
            return

        search_pair_modal = SearchPairModal(
            self._available_pairs,
            self._format_simple_pair,
            self,
        )
        search_pair_modal.pair_selected.connect(self.pair_changed.emit)
        search_pair_modal.show()

    def on_new_exchange(self, new_exchange: ExchangeBase) -> None:
        """Update info based on new exchange.

        Args:
            new_exchange (ExchangeBase): New exchangeBase.
        """
        self._current_pair = new_exchange.default_pair
        self._available_pairs = new_exchange.available_pairs
        self._format_simple_pair = new_exchange.format_simple_pair_from_pair


class SearchPairModal(QWidget):
    """Modal widget to search for available pairs."""

    pair_selected = Signal(str)

    def __init__(
        self,
        available_pairs: set[str],
        format_simple_pair_from_pair: Callable[[str], str],
        parent: Optional[QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(parent, Qt.WindowType.WindowStaysOnTopHint)
        self._available_pairs = {
            format_simple_pair_from_pair(pair): pair for pair in available_pairs
        }

        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

        self._completer = QCompleter(list(self._available_pairs.keys()))
        self._completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self._completer.setFilterMode(Qt.MatchFlag.MatchContains)

        self._main_layout = QVBoxLayout(self)
        self.search_input = VimLineEdit(self._completer)
        self.search_input.setFixedHeight(40)
        self.search_input.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        self.search_input.setPlaceholderText("Search for pair")
        self.search_input.setClearButtonEnabled(True)
        self.search_input.setCursor(Qt.CursorShape.ArrowCursor)
        self.search_input.returnPressed.connect(self.on_search_pair)

        self.parent().installEventFilter(self)

        self.close_shortcut = QShortcut(QKeySequence("Escape"), self)
        self.close_shortcut.activated.connect(self.close)

        self._main_layout.addWidget(self.search_input)
        self._main_layout.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.setAttribute(Qt.WidgetAttribute.WA_StyledBackground, True)
        self.setCursor(Qt.CursorShape.PointingHandCursor)

    def on_search_pair(self) -> None:
        """Emit pair selected signal if search input is valid."""
        if self.search_input.text() in self._available_pairs:
            self.pair_selected.emit(self._available_pairs[self.search_input.text()])
        self.close()

    def eventFilter(self, watched: QObject, event: QEvent) -> bool:
        """Filter resize events.

        Args:
            watched: The object being watched
            event: The event being watched

        Returns:
            bool: True if the event was handled.
        """
        if watched == self.parent() and event.type() == QEvent.Type.Resize:
            self.resize(self.parent().size())  # type: ignore
        return super().eventFilter(watched, event)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """Close modal when clicked outside of the image."""
        if not self.search_input.geometry().contains(event.pos()):
            self.close()
        return super().mousePressEvent(event)

    def showEvent(self, event: QShowEvent) -> None:
        """Raise and resize modal widget."""
        self.raise_()
        self.resize(self.parent().size())  # type: ignore
        self.search_input.setFocus()
        return super().showEvent(event)

    def closeEvent(self, event: QCloseEvent) -> None:
        """Delete widget on close."""
        self.deleteLater()
        return super().closeEvent(event)


class VimLineEdit(QLineEdit):
    """Line edit with vim motions for completer."""

    def __init__(self, completer: QCompleter, parent: Optional[QWidget] = None) -> None:
        """Initialize widget."""
        super().__init__(parent)
        self._completer = completer
        self.setCompleter(self._completer)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """Keypress event for Tab and Shift+Tab."""
        if event.key() == Qt.Key.Key_J and event.modifiers() & Qt.KeyboardModifier.ControlModifier:
            # Simulate down/up arrow key press to navigate the completer popup
            new_event = QKeyEvent(
                QEvent.Type.KeyPress,
                Qt.Key.Key_Down,
                Qt.KeyboardModifier.NoModifier,
            )
            QApplication.postEvent(self._completer.popup(), new_event)
        elif (
            event.key() == Qt.Key.Key_K and event.modifiers() & Qt.KeyboardModifier.ControlModifier
        ):
            # Simulate down/up arrow key press to navigate the completer popup
            new_event = QKeyEvent(
                QEvent.Type.KeyPress,
                Qt.Key.Key_Up,
                Qt.KeyboardModifier.NoModifier,
            )
            QApplication.postEvent(self._completer.popup(), new_event)
        elif event.key() in [Qt.Key.Key_Return, Qt.Key.Key_Enter]:
            if self._completer.popup().isVisible():
                current_index = self._completer.popup().currentIndex()
                # Select the item if there is one
                if not current_index.isValid() and self.text():
                    index = self._completer.completionModel().index(0, 0)
                    if index.isValid():
                        self.setText(self._completer.completionModel().data(index))
            super().keyPressEvent(event)
        else:
            super().keyPressEvent(event)



"""Useful log utilities."""

from collections.abc import Callable
from datetime import datetime, timezone
import logging.config
from pathlib import Path
import sys
from types import TracebackType
from typing import Any

from tenacity import RetryCallState

CURRENT_DIR = Path(__file__).parent
CONFIG_PATH = CURRENT_DIR.parent.joinpath("log_config.ini")


def setup_logging() -> None:
    """Set logging for the app based on config."""
    if not CONFIG_PATH.exists():
        create_default_config()

    log_path = Path(
        CONFIG_PATH.parent.joinpath("logs").joinpath(
            datetime.now(tz=timezone.utc).strftime("%Y-%m-%d_%H-%M-%S.log"),
        ),
    )
    logging.config.fileConfig(
        CONFIG_PATH,
        defaults={
            "log_path": f"{log_path.absolute()}",
        },
        disable_existing_loggers=False,
    )

    sys.excepthook = log_uncaught_exceptions


def create_default_config() -> None:
    """Create default logging config."""
    config_content = """[loggers]
keys=root,plutus_terminal

[handlers]
keys=consoleHandler, fileHandler

[logger_root]
level=ERROR
handlers=consoleHandler

[logger_plutus_terminal]
level=DEBUG
handlers=consoleHandler, fileHandler
qualname=plutus_terminal
propagate=0

[formatters]
keys=detailed, sampleFormatter

[formatter_detailed]
format=%(asctime)s [%(levelname)s] %(name)s: %(message)s

[handler_consoleHandler]
class=StreamHandler
level=INFO
formatter=detailed
args=(sys.stdout,)
encoding=utf-8

[handler_fileHandler]
class=FileHandler
level=DEBUG
formatter=sampleFormatter
args=(r"%(log_path)s", "a", "utf-8")

[formatter_sampleFormatter]
format=%(asctime)s - %(name)s - %(levelname)s - %(message)s
"""

    with Path.open(CONFIG_PATH, "w") as f:
        f.write(config_content)

    if not CONFIG_PATH.parent.joinpath("logs").exists():
        Path.mkdir(CONFIG_PATH.parent.joinpath("logs"), parents=True)


def log_uncaught_exceptions(
    exc_type: type[BaseException],
    exc_value: BaseException,
    exc_traceback: TracebackType | None,
) -> Any:  # noqa: ANN401
    """Log uncaught exceptions.

    Args:
        exc_type (type[BaseException]): Exception type.
        exc_value (BaseException): Exception value.
        exc_traceback (TracebackType): Exception traceback.
    """
    logger = logging.getLogger("plutus_terminal")
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    logger.critical("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))


def log_retry(logger: logging.Logger) -> Callable[[RetryCallState], None]:
    """Log warning after a retry happens.

    Returns:
        Callable[[RetryCallState], None]: Logging function.
    """

    def _log_retry(retry_state: RetryCallState) -> None:
        logger.warning(
            f"{retry_state.fn} failed after {retry_state.attempt_number} attempts"  # noqa: G004
            f"with exception: {retry_state.outcome.exception()}.",  # type: ignore
        )

    return _log_retry



"""Widget to manage trades on exchanges."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from PySide6 import QtWidgets
from PySide6.QtCore import Signal

from plutus_terminal.core.exchange.types import OrderData, PerpsPosition
from plutus_terminal.ui.widgets.orders_table import OrdersTableModel, OrdersTableView
from plutus_terminal.ui.widgets.positions_table import (
    PositionsTableModel,
    PositionsTableView,
)

if TYPE_CHECKING:
    from plutus_terminal.core.exchange.base import ExchangeBase


class TradeTable(QtWidgets.QWidget):
    """Widget to visualize and manage open trades."""

    pair_clicked = Signal(str)
    order_clicked = Signal(str)
    edit_order = Signal(OrderData)

    def __init__(
        self,
        exchange: ExchangeBase,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize shared viarables."""
        super().__init__(parent)

        self._main_layout = QtWidgets.QVBoxLayout()

        self._tab_widget = QtWidgets.QTabWidget()
        self._positions_model = PositionsTableModel(exchange.format_simple_pair_from_pair)
        self._positions_table = PositionsTableView(
            exchange,
        )
        self._orders_model = OrdersTableModel(exchange.format_simple_pair_from_pair)
        self._orders_table = OrdersTableView(exchange)

        self._setup_widgets()
        self._setup_layout()

        # Set minimum height to 10% of the widget height
        self.setMinimumHeight(int(self.sizeHint().height() * 1.1))

    def _setup_widgets(self) -> None:
        """Configure widgets."""
        self._positions_table.setModel(self._positions_model)
        self._positions_table.row_clicked.connect(self.pair_clicked)
        self._tab_widget.addTab(self._positions_table, "Positions (0)")

        self._orders_table.setModel(self._orders_model)
        self._tab_widget.addTab(self._orders_table, "Orders (0)")

    def _setup_layout(self) -> None:
        """Configure layout."""
        self._main_layout.addWidget(self._tab_widget)
        self.setLayout(self._main_layout)

    def update_positions(self, positions: list[PerpsPosition]) -> None:
        """Update positions."""
        self._tab_widget.setTabText(0, f"Positions ({len(positions)})")
        self._positions_model.update_positions(positions)

    def update_orders(self, orders: list[OrderData]) -> None:
        """Update orders."""
        self._tab_widget.setTabText(1, f"Orders ({len(orders)})")
        self._orders_model.update_orders(orders)

    def update_prices(self, cached_prices: dict) -> None:
        """Update prices."""
        self._positions_table.update_cached_prices(cached_prices)

    def on_new_exchange(self, new_exchange: ExchangeBase) -> None:
        """Update info based on new exchange.

        Args:
            new_exchange (ExchangeBase): New exchangeBase.
        """
        self._positions_model.on_new_exchange(new_exchange)
        self._orders_model.on_new_exchange(new_exchange)



"""Image web viewer widget."""

from __future__ import annotations

from typing import Optional

from PySide6.QtCore import QEvent, QObject, Qt, QUrl
from PySide6.QtGui import (
    QCloseEvent,
    QKeySequence,
    QMouseEvent,
    QPixmap,
    QShortcut,
    QShowEvent,
)
from PySide6.QtNetwork import QNetworkAccessManager, QNetworkReply, QNetworkRequest
from PySide6.QtWidgets import QApplication, QLabel, QMainWindow, QVBoxLayout, QWidget


class ImageWebViewer(QLabel):
    """Display images from web that can be zoomed in a modal widget."""

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        """Initialize widget shared variables."""
        super().__init__(parent=parent)
        self._pixmap = QPixmap()
        self._max_img_height = 180
        self.setCursor(Qt.CursorShape.PointingHandCursor)

    def set_image(self, image_url: str) -> None:
        """Set image from url."""
        pixmap = QPixmap()
        newtwork_manager = QNetworkAccessManager(self)
        newtwork_manager.finished.connect(self._set_newtwork_image)
        newtwork_manager.get(QNetworkRequest(QUrl(image_url)))

        self.setPixmap(pixmap)

    def _set_newtwork_image(self, reply: QNetworkReply) -> None:
        """Set image pixmap from network reply."""
        image_data = reply.readAll()
        pixmap = QPixmap()
        self._pixmap.loadFromData(image_data)
        height = (
            self._pixmap.height()
            if self._pixmap.height() < self._max_img_height
            else self._max_img_height
        )
        pixmap = self._pixmap.scaled(
            self.parentWidget().size().width(),
            height,
            Qt.AspectRatioMode.KeepAspectRatio,
            Qt.TransformationMode.SmoothTransformation,
        )
        self.setPixmap(pixmap)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """Open model for image when clicked."""
        ImageDisplayModal.show_modal(self._pixmap)
        return super().mousePressEvent(event)


class ImageDisplayModal(QWidget):
    """Display image modal widget."""

    def __init__(self, parent: QWidget) -> None:
        """Initialize modal widget."""
        super().__init__(parent, Qt.WindowType.WindowStaysOnTopHint)

        self._main_layout = QVBoxLayout(self)
        self._image = QLabel()
        self._image.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._image.setCursor(Qt.CursorShape.ArrowCursor)

        self.parent().installEventFilter(self)

        self.close_shortcut = QShortcut(QKeySequence("Escape"), self)
        self.close_shortcut.activated.connect(self.close)

        self._main_layout.addWidget(self._image)
        self._main_layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setAttribute(Qt.WidgetAttribute.WA_StyledBackground, True)
        self.setCursor(Qt.CursorShape.PointingHandCursor)

    def set_image(self, image_pixmap: QPixmap) -> None:
        """Set image from url."""
        self.resize_image(image_pixmap)

    def resize_image(self, pixmap: QPixmap) -> None:
        """Resize image based on the current widget size."""
        max_size = self.parent().size() * 0.7  # type: ignore
        image_size = max_size if pixmap.height() > max_size.height() else pixmap.size()
        scaled_pixmap = pixmap.scaled(image_size, Qt.AspectRatioMode.KeepAspectRatio)
        self._image.setPixmap(scaled_pixmap)

    def eventFilter(self, watched: QObject, event: QEvent) -> bool:
        """Filter resize events.

        Args:
            watched: The object being watched
            event: The event being watched

        Returns:
            bool: True if the event was handled.
        """
        if watched == self.parent() and event.type() == QEvent.Type.Resize:
            self.resize_image(self._image.pixmap())
            self.resize(self.parent().size())  # type: ignore
        return super().eventFilter(watched, event)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """Close modal when clicked outside of the image."""
        if not self._image.geometry().contains(event.pos()):
            self.close()
        return super().mousePressEvent(event)

    def showEvent(self, event: QShowEvent) -> None:
        """Raise and resize modal widget."""
        self.raise_()
        self.resize(self.parent().size())  # type: ignore
        return super().showEvent(event)

    def closeEvent(self, event: QCloseEvent) -> None:
        """Delete widget on close."""
        self.deleteLater()
        return super().closeEvent(event)

    @staticmethod
    def show_modal(image_pixmap: QPixmap) -> None:
        """Show modal widget."""
        main_window = None
        for widget in QApplication.topLevelWidgets():
            if isinstance(widget, QMainWindow):
                main_window = widget
                break
        if not main_window:
            return
        parent = main_window
        image_display = ImageDisplayModal(parent)
        image_display.set_image(image_pixmap)
        image_display.show()



"""Types for filters."""

from __future__ import annotations

from enum import IntEnum


class FilterType(IntEnum):
    """Filter types."""

    KEYWORD_MATCHING = 0
    DATA_MATCHING = 1


class ActionType(IntEnum):
    """Filter Action types."""

    COIN_ASSOCIATION = 0
    SOUND_ASSOCIATION = 1
    IGNORE = 2



"""Foxify exchange options."""

from __future__ import annotations

import base64
import binascii
import logging
from typing import TYPE_CHECKING, Self

from httpx import AsyncClient, Client, ConnectError, HTTPStatusError, ReadTimeout
import pandas
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from web3 import AsyncHTTPProvider, AsyncWeb3
from web3.exceptions import ContractLogicError
from web3.types import Gwei, Nonce, Wei

from plutus_terminal.core.config import CONFIG
from plutus_terminal.core.exceptions import TransactionFailedError
from plutus_terminal.core.exchange.base import ExchangeOptions
from plutus_terminal.core.exchange.foxify import utils as foxify_utils
from plutus_terminal.core.exchange.web3 import web3_utils
from plutus_terminal.core.types_ import (
    OptionsBuyParams,
    OptionsDirection,
    OptionsOrdersParams,
    OptionsPercent,
    OptionsRisk,
    OptionsSortingBy,
    OptionsStrategy,
)
from plutus_terminal.log_utils import log_retry

if TYPE_CHECKING:
    from decimal import Decimal

    from eth_account.signers.local import LocalAccount


LOGGER = logging.getLogger(__name__)


class FoxifyOptions(ExchangeOptions):
    """Handle Options on Foxify Exchange."""

    def __init__(self, web3_account: LocalAccount, extra_gas: Gwei) -> None:
        """Initialize shared attributes.

        Args:
            web3_account (LocalAccount): Web3 Account.
            extra_gas (Gwei): Extra gas for all transactions.
        """
        self.web3_account = web3_account
        self.extra_gas = extra_gas
        self.web3_provider = AsyncWeb3(
            AsyncHTTPProvider(str(CONFIG.get_web3_rpc_by_name("Arbitrum One").rpc_url)),
        )
        self.core_contract = foxify_utils.build_options_core_contract(
            self.web3_provider,
        )

    @classmethod
    async def create(cls, web3_account: LocalAccount, extra_gas: Gwei) -> Self:
        """Create class instance and init_async.

        Args:
            web3_account (LocalAccount): Web3 Account.
            extra_gas (Gwei): Extra gas for all transactions.

        Returns:
            FoxifyOptions: Instance of FoxifyOptions.
        """
        return cls(web3_account, extra_gas)

    async def fetch_orders(
        self,
        options_orders_params: OptionsOrdersParams,
    ) -> pandas.DataFrame:
        """Fetch options orders based on parameters.

        Args:
            options_orders_params (OrdersParams): Params for request.

        Returns:
            pandas.DataFrame: DataFrame with options orders.
        """
        request_url = "https://api-options.prd.foxify.trade/api/orders"
        async with AsyncClient() as client:
            response = await client.get(
                request_url,
                params=options_orders_params,  # type: ignore
            )
        response.raise_for_status()

        data = pandas.DataFrame(response.json())
        data["id"] = data["id"].astype(int)
        data["orderId"] = data["orderId"].astype(int)
        data["amount"] = data["amount"].astype("int64")
        data["reserved"] = data["reserved"].astype("int64")
        data["available"] = data["available"].astype("int64")
        data["percent"] = data["percent"].astype("int64")
        data["rate"] = data["rate"].astype("int64")
        data["duration"] = data["duration"].astype("int64")
        return data

    @retry(
        retry=(
            retry_if_exception_type(HTTPStatusError)
            | retry_if_exception_type(ReadTimeout)
            | retry_if_exception_type(ConnectError)
        ),
        reraise=True,
        stop=stop_after_attempt(5),
        wait=wait_exponential(multiplier=1, min=0.15, max=1),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def fetch_latest_vaa(self, price_feed_id: str) -> str:
        """Fetch the latest price vaa from Pyth using Hermes.

        Args:
            price_feed_id (str): Price feed ID to fetch vaa.

        Returns:
            str: VVA for the price.

        Raises:
            HTTPStatusError: If response is not successful. After tries.
        """
        request_url = "https://hermes.pyth.network/api/latest_vaas"
        request_params = {"ids[]": [price_feed_id]}

        async with AsyncClient() as client:
            response = await client.get(request_url, params=request_params)
        response.raise_for_status()

        data = response.json()
        return "0x" + binascii.hexlify(base64.b64decode(data[0])).decode("ascii")

    @retry(
        retry=(
            retry_if_exception_type(HTTPStatusError)
            | retry_if_exception_type(ReadTimeout)
            | retry_if_exception_type(ConnectError)
        ),
        reraise=True,
        stop=stop_after_attempt(5),
        wait=wait_exponential(multiplier=1, min=0.15, max=1),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def _notify_api_options_bought(self, tx_hash: str) -> None:
        """Notify API that an order was bought.

        Args:
            tx_hash (str): Tx hash to notify.
        """
        request_url = "https://api-options.prd.foxify.trade/api/positions/accept"
        request_params = {"txHash": tx_hash}

        async with AsyncClient() as client:
            response = await client.post(request_url, json=request_params)
        response.raise_for_status()

    async def approve_stable(self) -> None:
        """Approve stable."""
        await foxify_utils.approve_stable(
            self.web3_provider,
            self.core_contract,
            self.web3_account,
        )

    async def buy_options(self, options_buy_params: OptionsBuyParams) -> None:
        """Buy options.

        Args:
            options_buy_params (OptionsBuyParams): Params to buy options.

        Raises:
            TransactionError: If transaction fails.
        """
        nonce: Nonce = await self.web3_provider.eth.get_transaction_count(
            self.web3_account.address,
        )
        latest_vaa = await self.fetch_latest_vaa(options_buy_params["price_id"])
        data = []
        for order_id, amount in options_buy_params["orders"].items():
            data.append(
                {
                    "orderId": order_id,
                    "amount": amount,
                    "updateData": [latest_vaa],
                },
            )

        try:
            tx = await self.core_contract.functions.accept(
                self.web3_account.address,
                data,
            ).build_transaction(
                {
                    "nonce": nonce,
                    "from": self.web3_account.address,
                    "maxFeePerGas": await web3_utils.estimate_gas_price(
                        self.web3_provider,
                        self.extra_gas,
                    ),
                    "maxPriorityFeePerGas": self.web3_provider.to_wei(
                        self.extra_gas,
                        "gwei",
                    ),
                    "gas": Wei(1000000),
                },
            )
            tx.update({"gas": await web3_utils.estimate_gas(self.web3_provider, tx)})
        except (ContractLogicError, ValueError) as error:
            raise TransactionFailedError from error

        signed_tx = self.web3_account.sign_transaction(tx)
        send_txn = await self.web3_provider.eth.send_raw_transaction(
            signed_tx.rawTransaction,
        )
        await self.web3_provider.eth.wait_for_transaction_receipt(send_txn)
        await self._notify_api_options_bought(self.web3_provider.to_hex(send_txn))
        LOGGER.info(
            "Bought options tx: https://arbiscan.io/tx/%s",
            self.web3_provider.to_hex(send_txn),
        )

    def fetch_available_pairs(self) -> list[str]:
        """Fetch available pairs for options.

        Returns:
            list[str]: List of available pairs.
        """
        request_url = "https://api-options.prd.foxify.trade/api/oracles"

        with Client() as client:
            response = client.get(request_url)
        response.raise_for_status()
        data = response.json()

        return [entry["name"] for entry in data]

    async def filter_with_strategy(
        self,
        direction: OptionsDirection,
        value: Decimal,
        pair: str,
    ) -> pandas.DataFrame:
        """Filter options based on strategy.

        Args:
            direction (OptionsDirection): Direction to buy the option.
            value (Decimal): Max value to spend.
            pair (str): Pair to open options for.

        Returns:
            pandas.DataFrame: DataFrame with options orders.
        """
        # build strategy
        percent_prefix = "UP" if direction == OptionsDirection.UP else "DOWN"

        percent_min_key = f"{percent_prefix}_{CONFIG.options_percent_min.replace('.', '')[:-1]}"
        percent_max_key = f"{percent_prefix}_{CONFIG.options_percent_max.replace('.', '')[:-1]}"

        percent_min = OptionsPercent[
            percent_min_key if CONFIG.options_percent_min != "0%" else percent_prefix
        ]
        percent_max = OptionsPercent[
            percent_max_key if CONFIG.options_percent_max != "0%" else percent_prefix
        ]

        strategy = OptionsStrategy(
            {
                "direction": direction,
                "rate_min": CONFIG.options_rate_min,
                "available_min": CONFIG.options_available_min,
                "percent_min": percent_min,
                "percent_max": percent_max,
                "duration_min": CONFIG.options_duration_min,
                "duration_max": CONFIG.options_duration_max,
                "risk": CONFIG.options_risk,
            },
        )

        # fetch current orders
        options_params: OptionsOrdersParams = {
            "sorting_by": OptionsSortingBy.RATE,
            "closed": False,
        }
        orders_data = await self.fetch_orders(options_params)

        # Common filtering logic
        common_filter = (
            (orders_data["rate"] + (1 * 10**18) >= strategy["rate_min"] * 10**18)
            & (orders_data["available"] >= strategy["available_min"] * 10**6)
            & (orders_data["direction"] == str(strategy["direction"]))
            & (orders_data["duration"] >= strategy["duration_min"])
            & (orders_data["duration"] <= strategy["duration_max"])
            & (orders_data["creator"] != self.web3_account.address.lower())
            & (orders_data["oracle"].apply(lambda x: x.get("name")) == pair)
        )

        # Direction-specific filtering
        if direction == OptionsDirection.UP:
            direction_filter = (orders_data["percent"] >= strategy["percent_min"]) & (
                orders_data["percent"] <= strategy["percent_max"]
            )

        else:
            direction_filter = (orders_data["percent"] <= strategy["percent_min"]) & (
                orders_data["percent"] >= strategy["percent_max"]
            )

        filtered_data = orders_data[common_filter & direction_filter]

        # Apply risk strategy to filtered order
        sort_by, ascending = [], []
        if strategy["risk"] == OptionsRisk.PRIO_RETURN:
            sort_by = ["rate", "available"]
            ascending = [False, False]
        elif strategy["risk"] == OptionsRisk.PRIO_SAFETY:
            sort_by = ["duration", "percent", "rate", "available"]
            ascending = [True, True, False, False]

        if sort_by:
            filtered_data = filtered_data.sort_values(by=sort_by, ascending=ascending)
        filtered_data = filtered_data.reset_index(drop=True)

        if filtered_data.empty:
            LOGGER.debug("No options to buy, based on strategy: %s", strategy)
            return pandas.DataFrame()

        # Process for target value
        filtered_data["cumulative_available"] = filtered_data["available"].cumsum()
        max_cumulative = filtered_data["cumulative_available"].iloc[-1]
        target_value = int(value.scaleb(6))
        target_sum = min(target_value, max_cumulative)

        last_index = (
            filtered_data[filtered_data["cumulative_available"] > target_sum].index[0]
            if target_value <= max_cumulative
            else filtered_data.index[-1]
        )
        return filtered_data.loc[:last_index]



"""Run plutus terminal."""

import asyncio
import gc
from pathlib import Path
import platform
import sys

from PySide6.QtGui import QPixmap
from PySide6.QtWidgets import (
    QApplication,
    QMenu,
    QSplashScreen,
    QSystemTrayIcon,
)
from qasync import QEventLoop, asyncSlot

from plutus_terminal.core.config import CONFIG, AppConfig
from plutus_terminal.log_utils import setup_logging
from plutus_terminal.ui import resources
from plutus_terminal.ui.main_window import PlutusTerminal
from plutus_terminal.ui.widgets.new_account import NewAccountDialog


class PlutusSystemTrayApp(QApplication):
    """Plutus System Tray App."""

    def __init__(self, argv: list[str]) -> None:
        """Initialize."""
        super().__init__(argv)
        self.splash_screen = QSplashScreen()
        self.splash_screen.setPixmap(QPixmap(":/general/splash_screen"))
        self.splash_screen.show()
        self.processEvents()

        self.main_window = PlutusTerminal()

        self._tray_icon = QSystemTrayIcon()
        self._tray_icon.setIcon(QPixmap(":/icons/plutus_icon"))
        self._tray_icon.setToolTip("Plutus Terminal")

        self._init_tray()

    def _init_tray(self) -> None:
        """Initialize tray icon."""
        menu = QMenu()
        menu.addAction("Open Terminal", self.main_window.show)
        menu.addAction("Exit", self.exit)

        self._tray_icon.setContextMenu(menu)
        self._tray_icon.show()

        self._tray_icon.activated.connect(self._on_tray_activated)

    def _on_tray_activated(self, reason: int) -> None:
        """Handle tray icon activation."""
        if reason == QSystemTrayIcon.ActivationReason.DoubleClick:
            self.main_window.show()

    async def init_and_show(self) -> None:
        """Initialize window and show."""
        CONFIG.load_config()
        await self.main_window.init_async()
        self.splash_screen.hide()
        self.main_window.show()

    def validate_if_account(self) -> None:
        """Validate if there is at least one account.

        If no account is found a new account dialog will be shown.
        """
        if not AppConfig.get_all_keyring_accounts():
            new_account_dialog = NewAccountDialog()
            if not new_account_dialog.exec():
                sys.exit()

    @asyncSlot()
    async def cleanup(self) -> None:
        """Clean up async connections before closing."""
        await self.main_window.stop_async()


def run() -> None:
    """Run plutus terminal."""
    # Override gc threshold
    gc.set_threshold(100_000, 50, 100)

    # Set process name
    if platform.system() == "Windows":
        import ctypes

        ctypes.windll.kernel32.SetConsoleTitleW("Plutus Terminal")  # type: ignore
    else:
        import setproctitle

        setproctitle.setproctitle("Plutues Terminal")

    setup_logging()
    app = PlutusSystemTrayApp([])
    relative_path = Path(__file__).parent
    with Path.open(relative_path.joinpath("ui/style.qss")) as f:
        app.setStyleSheet(f.read())

    event_loop = QEventLoop(app)
    asyncio.set_event_loop(event_loop)

    app_close_event = asyncio.Event()
    app.aboutToQuit.connect(app_close_event.set)

    app.validate_if_account()

    async def run_and_await() -> None:
        """Run all and await close event."""
        await app.init_and_show()
        await app_close_event.wait()

    event_loop.create_task(run_and_await())
    event_loop.run_forever()

    event_loop.run_until_complete(app.cleanup())
    event_loop.close()


if __name__ == "__main__":
    run()



"""New list widget."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from PySide6 import QtWidgets
from PySide6.QtCore import Qt, QUrl, Signal
from PySide6.QtGui import QKeySequence, QPixmap, QShortcut
from PySide6.QtMultimedia import QSoundEffect

from plutus_terminal.core.config import CONFIG
from plutus_terminal.ui.ui_utils import list_resources_from_prefix
from plutus_terminal.ui.widgets.news_widget import NewsWidget
from plutus_terminal.ui.widgets.toast import Toast
from plutus_terminal.ui.widgets.top_bar_widget import TopBar

if TYPE_CHECKING:
    from plutus_terminal.core.exchange.base import ExchangeBase
    from plutus_terminal.core.types_ import NewsData


class NewsList(QtWidgets.QWidget):
    """News list widget.

    Acts like a news factory and handles news items.
    """

    pair_clicked = Signal(str)
    refresh_news = Signal()

    def __init__(
        self,
        exchange: ExchangeBase,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize shared attributes."""
        super().__init__(parent=parent)
        self._exchange = exchange

        self._main_layout = QtWidgets.QVBoxLayout()

        self.top_bar = TopBar("News Feed")
        self._top_bar_show_images = QtWidgets.QRadioButton()
        self._top_bar_notifications = QtWidgets.QRadioButton()
        self._top_bar_max_news = QtWidgets.QComboBox()

        self._scroll_area = QtWidgets.QScrollArea()
        self._content_area = QtWidgets.QWidget()
        self._scroll_layout = QtWidgets.QVBoxLayout()

        self._sfxs: dict[str, QSoundEffect] = {}

        self._selected_news_widget: Optional[NewsWidget] = None

        self.max_news = 25

        self._load_sfxs()
        self._setup_widgets()
        self._setup_layout()
        self._setup_shorcust()
        self._show_widget_index_at_top(0)

    def _load_sfxs(self) -> None:
        """Load sound effects in memory."""
        for sfx_name in list_resources_from_prefix("sfx"):
            sfx_path = f":/sfx/{sfx_name}"
            sfx = QSoundEffect()
            sfx.setSource(QUrl.fromLocalFile(sfx_path))
            self._sfxs[sfx_path] = sfx

    def _setup_widgets(self) -> None:
        """Create internal widgets."""
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Fixed,
            QtWidgets.QSizePolicy.Policy.MinimumExpanding,
        )

        self.top_bar.icon.setPixmap(QPixmap(":/icons/news_feed_icon"))

        self._top_bar_show_images.setAutoExclusive(False)
        self._top_bar_show_images.setIcon(QPixmap(":/icons/gallery_icon"))
        self._top_bar_show_images.setChecked(
            CONFIG.get_gui_settings("news_show_images"),
        )
        self._top_bar_show_images.setToolTip("Show Images")
        self._top_bar_show_images.toggled.connect(self.show_images_toggled)

        self._top_bar_notifications.setAutoExclusive(False)
        self._top_bar_notifications.setChecked(
            CONFIG.get_gui_settings("news_desktop_notifications"),
        )
        if self._top_bar_notifications.isChecked():
            self._top_bar_notifications.setIcon(QPixmap(":/icons/notification_on"))
        else:
            self._top_bar_notifications.setIcon(QPixmap(":/icons/notification_off"))
        self._top_bar_notifications.setToolTip("Enable Desktop Notifications")
        self._top_bar_notifications.toggled.connect(self.notifications_toggled)

        self._top_bar_max_news.addItems(["25 results", "50 results", "100 results", "200 results"])
        self._top_bar_max_news.currentIndexChanged.connect(
            self.update_max_news,
        )

        self.top_bar.add_widget(self._top_bar_show_images)
        self.top_bar.add_widget(self._top_bar_notifications)
        self.top_bar.add_widget(self._top_bar_max_news)

        self._scroll_area.setWidgetResizable(True)
        self._scroll_area.setVerticalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOff,
        )

        self._scroll_area.verticalScrollBar().rangeChanged.connect(
            lambda: self._show_widget_at_top(self._selected_news_widget),
        )

    def _setup_layout(self) -> None:
        """Configure layouts."""
        self._main_layout.addWidget(self.top_bar)
        self._scroll_area.setWidget(self._content_area)
        self._content_area.setLayout(self._scroll_layout)
        self._main_layout.addWidget(self._scroll_area)

        self.setLayout(self._main_layout)

    def _setup_shorcust(self) -> None:
        """Configure shortcuts."""
        self._reset_scroll_shortcut = QShortcut(QKeySequence("q"), self)
        self._reset_scroll_shortcut.activated.connect(self._reset_scroll_to_top)

        self._step_up_shortcut = QShortcut(QKeySequence("w"), self)
        self._step_up_shortcut.activated.connect(
            self._step_widget_up,
        )

        self._set_down_shortcut = QShortcut(QKeySequence("s"), self)
        self._set_down_shortcut.activated.connect(self._step_widget_down)

        self._open_link_shortcut = QShortcut(QKeySequence("space"), self)
        self._open_link_shortcut.activated.connect(self._open_link)

    def _reset_scroll_to_top(self) -> None:
        """Reset the scroll position to 0."""
        if self._selected_news_widget is None:
            self._selected_news_widget = self._scroll_layout.itemAt(0).widget()  # type: ignore
        self._show_widget_index_at_top(0)

    def _step_widget_up(self) -> None:
        """Move scroll bar 1 news widget up."""
        if self._selected_news_widget is None:
            self._selected_news_widget = self._scroll_layout.itemAt(0).widget()  # type: ignore
        current_index = self._scroll_layout.indexOf(self._selected_news_widget)  # type: ignore
        self._show_widget_index_at_top(current_index - 1)

    def _step_widget_down(self) -> None:
        """Move scroll bar 1 news widget down."""
        if self._selected_news_widget is None:
            self._selected_news_widget = self._scroll_layout.itemAt(0).widget()  # type: ignore
        current_index = self._scroll_layout.indexOf(self._selected_news_widget)  # type: ignore
        self._show_widget_index_at_top(current_index + 1)

    def _show_widget_index_at_top(self, index: int) -> None:
        """Move scroll area to show widget at the top."""
        # Ensure the index is within valid range
        if index < 0 or index >= self._scroll_layout.count():
            return

        news_widget = self._scroll_layout.itemAt(index).widget()
        if not isinstance(news_widget, NewsWidget):
            return
        self._select_news_widget(news_widget)

        # Calculate the cumulative height of all widgets up to the selected one
        cumulative_height = 0
        for i in range(index):
            widget = self._scroll_layout.itemAt(i).widget()
            if widget.isVisible():  # Consider only visible widgets
                cumulative_height += widget.height()  # Use actual height

            # Add spacing after each widget except the last one
            if i < index - 1:
                cumulative_height += self._scroll_layout.spacing()

        # Include top margin of the layout
        cumulative_height += self._scroll_layout.contentsMargins().top()

        # Set the vertical scroll bar's value to the cumulative height
        self._scroll_area.verticalScrollBar().setValue(cumulative_height)

    def _show_widget_at_top(self, news_widget: Optional[NewsWidget]) -> None:
        """Move scroll area to show widget at the top."""
        if news_widget is None:
            return
        self._show_widget_index_at_top(self._scroll_layout.indexOf(news_widget))

    def _select_news_widget(self, news_widget: NewsWidget) -> None:
        """Select news widget."""
        if news_widget is None:
            return

        # Ensure there is news selected
        if self._selected_news_widget is not None:
            self._selected_news_widget.set_unselected_style()

        # Select new widget
        self._selected_news_widget = news_widget
        self._selected_news_widget.set_selected_style()  # type: ignore

    def add_news(self, news_data: NewsData) -> None:
        """Add NewsWidget to the list and desktop notifications."""
        # Do not add ignored news
        if news_data["ignored"]:
            return

        self._sfxs[news_data["sfx"]].play()
        if CONFIG.get_gui_settings("news_desktop_notifications"):
            desktop_news = self._create_news_widget(news_data)
            Toast.show_widget(
                desktop_news,
                timeout=35000,
                desktop=True,
            )

        self._add_news_to_list(news_data)

    def fill_old_news(self, list_news: list[NewsData]) -> None:
        """Clear and fill list with given data."""
        self.setDisabled(True)
        self._scroll_area.blockSignals(True)
        self.clear_list()
        for news_data in list_news:
            # Do not add ignored news
            if news_data["ignored"]:
                continue

            self._add_news_to_list(news_data)
        self._scroll_area.blockSignals(False)
        self.setDisabled(False)

    def _add_news_to_list(self, news_data: NewsData) -> NewsWidget:
        """Add news to list respecting the limit.

        Args:
            news_data (NewsData): News data to create new widget.

        Returns:
            NewsWidget: Created news widget.
        """
        news_widget = self._create_news_widget(news_data)

        # If no news is selected, select the new one
        if self._selected_news_widget is None:
            self._selected_news_widget = news_widget
        # If the news is already selected, unselect it and select the new one
        elif self._scroll_layout.indexOf(self._selected_news_widget) == 0:
            self._selected_news_widget.set_unselected_style()
            self._selected_news_widget = news_widget

        # Remove oldest widget if the limit is reached
        self._scroll_area.blockSignals(True)
        if self._scroll_layout.count() > self.max_news:
            old_widget = self._scroll_layout.takeAt(
                self._scroll_layout.count() - 1,
            ).widget()
            if old_widget == self._selected_news_widget:
                self._selected_news_widget = self._scroll_layout.itemAt(
                    self.max_news - 1,
                ).widget()  # type: ignore
            old_widget.deleteLater()
        self._scroll_area.blockSignals(False)

        self._scroll_layout.insertWidget(0, news_widget)

        return news_widget

    def _create_news_widget(self, news_data: NewsData) -> NewsWidget:
        """Create news widget."""
        news_widget = NewsWidget(
            news_data,
            self._exchange.format_pair_from_coin,
            self._exchange.available_pairs,
        )
        news_widget.show_images = CONFIG.get_gui_settings("news_show_images")
        news_widget.create_interactions(self._exchange)
        news_widget.pair_clicked.connect(self.pair_clicked.emit)
        news_widget.news_clicked.connect(self._show_widget_at_top)
        return news_widget

    def clear_list(self) -> None:
        """Clear list."""
        while self._scroll_layout.count():
            old_widget = self._scroll_layout.takeAt(
                self._scroll_layout.count() - 1,
            ).widget()
            old_widget.deleteLater()
        self._selected_news_widget = None

    def _open_link(self) -> None:
        """Open link of selected news in browser."""
        if self._selected_news_widget is None:
            return
        self._selected_news_widget.open_link()

    def show_images_toggled(self, value: bool) -> None:
        """Show images toggled."""
        CONFIG.set_gui_settings("news_show_images", value)
        for index in range(self._scroll_layout.count()):
            widget = self._scroll_layout.itemAt(index).widget()
            if isinstance(widget, NewsWidget):
                widget.show_images = value

    def update_news_trade_buttons(self) -> None:
        """Update trade values for all NewsWidgets buttons."""
        for index in range(self._scroll_layout.count()):
            widget = self._scroll_layout.itemAt(index).widget()
            if isinstance(widget, NewsWidget):
                widget.update_trade_buttons()

    def notifications_toggled(self, value: bool) -> None:
        """Notifications toggled."""
        CONFIG.set_gui_settings("news_desktop_notifications", value)
        if value:
            self._top_bar_notifications.setIcon(QPixmap(":/icons/notification_on"))
        else:
            self._top_bar_notifications.setIcon(QPixmap(":/icons/notification_off"))

    def on_new_exchange(self, exchange: ExchangeBase) -> None:
        """Update info based on new exchange.

        Args:
            exchange (ExchangeBase): New exchangeBase.
        """
        self._exchange = exchange
        self._selected_news_widget = None

    def update_max_news(self, max_news_index: int) -> None:
        """Update max news."""
        max_news_text = self._top_bar_max_news.itemText(max_news_index)
        max_news = int(max_news_text.split(" ")[0])
        self.max_news = max_news
        self.refresh_news.emit()



"""Foxify funded exchange fetcher."""

from decimal import Decimal
import logging

from eth_typing import ChecksumAddress
from tenacity import before_sleep_log, retry, wait_exponential

from plutus_terminal.core.exchange.base import ExchangeFetcherMessageBus
from plutus_terminal.core.exchange.foxify import utils as foxify_utils
from plutus_terminal.core.exchange.foxify.fetcher import FoxifyFetcher
from plutus_terminal.log_utils import log_retry

LOGGER = logging.getLogger(__name__)


class FoxifyFundedFetcher(FoxifyFetcher):
    """Foxify funded exchange fetcher."""

    def __init__(
        self,
        pair_map: dict[str, str],
        web3_address: ChecksumAddress,
        message_bus: ExchangeFetcherMessageBus,
    ) -> None:
        """Initialize shared attributes.

        Args:
            pair_map (dict[str, dict[str, str]]): Dict with adress and pair.
            web3_address (ChecksumAddress): Web3 account address.
            message_bus (ExchangeFetcherMessageBus): Message bus.
        """
        super().__init__(pair_map, web3_address, message_bus)
        self._funded_trader_contract = foxify_utils.build_funded_trader_contract(
            web3_address,
            self.web3_provider,
        )

    @retry(
        wait=wait_exponential(multiplier=1, min=0.15, max=5),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def fetch_stable_balance(self) -> Decimal:
        """Fetch stable balance.

        Returns:
            Decimal: Balance in USD Stable Format.
        """
        balance = await self._funded_trader_contract.functions.balance().call()

        return Decimal(balance) / 10**foxify_utils.USDC_DECIMAL_PLACES



"""CCXT Utilities."""

from __future__ import annotations

from typing import TYPE_CHECKING, NotRequired, TypedDict

if TYPE_CHECKING:
    from decimal import Decimal

    from plutus_terminal.core.types_ import PerpsTradeDirection, PerpsTradeType


class TradingArgs(TypedDict):
    """Args for trading with CCXT."""

    pair: str
    trade_type: PerpsTradeType
    side: PerpsTradeDirection
    amount: Decimal
    price: Decimal
    take_profit: NotRequired[Decimal]
    stop_loss: NotRequired[Decimal]
